/**
  *  \file client/chartloc.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chartloc.h"
#include "game/univ.h"
#include "game/global.h"
#include "game/ship.h"

GChartLocation::GChartLocation()
    : location(0, 0),
      wrap_adjust(0, 0),
      pid(0),
      list_index(0),
      changed(false)
{
    viewport.center.x = viewport.center.y = 0;
}

GChartLocation::~GChartLocation()
{ }

void
GChartLocation::updateLocation()
{
    GPoint realLocation = location.getNormalized();

    // Build object lists
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    pid = univ.getPlanetAt(realLocation);
    list.clear();
    list.addObjectsAt(univ, realLocation, ls_IncludeForeignShips + ls_IncludePlanet, 0);
    list.postprocessUser(false);

    // Reset list index
    list_index = (list.getNumObjects() > 0 ? 1 : 0);

    // Update object identifier
    updateObject();
}

void
GChartLocation::updateObject()
{
    GObjectId obj = getCurrentObject();
    if (obj.isValid()) {
        if (GShip* sh = dynamic_cast<GShip*>(&obj.getObject()))
            viewport.vec_id = sh->getId();
        else
            viewport.vec_id = 0;
    }
}

inline void
GChartLocation::clearWrapAdjust()
{
    wrap_adjust.x = wrap_adjust.y = 0;
}

/** Initialize before first load. */
void
GChartLocation::init()
{
    viewport.center.x = viewport.center.y = 2000;
    location.x = location.y = 2000;
    list.clear();
    clearWrapAdjust();
    pid = 0;
    changed = true;
}

/** Initialize after game load. */
void
GChartLocation::initAfterLoad()
{
    init();
    updateLocation();
}

/** Initialize upon starchart entry. */
void
GChartLocation::initEntry(const GfxRect& area)
{
    /* initialize viewport */
    viewport.mult = viewport.divi = 1;
    viewport.view = area;
    viewport.vec_id = 0;
    viewport.drawing_filter = viewport.NO_FILTER;
    viewport.init();
    updateObject();
    clearWrapAdjust();

    /* cause initial redraw */
    changed = true;
}

/** Set center of chart. */
void
GChartLocation::setCenter(GPoint pt)
{
    pt = pt.getNormalized();

    // Same as last time?
    if (pt == viewport.center)
        return;

    // Change center
    viewport.center = pt;
    changed = true;

    // Update location, it might jump over a map border now
    GPoint near = location.getNearestTo(pt);
    if (near != location) {
        wrap_adjust += (near - location);
        location = near;
    }
}

/** Set location. */
void
GChartLocation::setLocation(GPoint pt)
{
    pt = pt.getNormalized().getNearestTo(viewport.center);

    // Same as last time?
    if (pt == location)
        return;

    // Change location
    location = pt;
    changed = true;
    updateLocation();
}

/** Get cumulative wrap adjustment. If \c location has changed unsuspectedly
    due to a change of the center, the difference will be accumulated here. */
GPoint
GChartLocation::getWrapAdjust() const
{
    return wrap_adjust;
}

/** Get currently-locked object. */
GObjectId
GChartLocation::getCurrentObject() const
{
    if (list_index > 0 && list_index <= list.getNumObjects())
        return list.getObjectId(list_index); /* 1-based access! */
    else
        return GObjectId();
}

/** Set currently-locked object. Also adjusts the position if required. */
void
GChartLocation::setCurrentObject(GObject& obj)
{
    // If it's a map object, go there
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&obj)) {
        setLocation(mo->getPos());
    }

    // Focus it.
    int index = list.getIndexFor(obj);
    if (index > 0 && index != list_index) {
        list_index = index;
        updateObject();
        changed = true;
    }
}

/** Get Id of currently-locked ship, if any.
    \return ship id, 0 if no ship locked */
int
GChartLocation::getCurrentShipId() const
{
    GObjectId id = getCurrentObject();
    if (!id.isGameObject())
        return 0;

    GShip* sh = dynamic_cast<GShip*>(&id.getObject());
    if (!sh)
        return 0;

    return sh->getId();
}

/** Perform zooming. */
void
GChartLocation::doZoom(bool enlarge)
{
    if (enlarge)
        changed |= viewport.zoomIn(getDisplayedTurn().getChartBBox());
    else
        changed |= viewport.zoomOut(getDisplayedTurn().getChartBBox());
}

/** Scroll through objects at current location. */
void
GChartLocation::doScroll(int direction, bool mark)
{
    int new_index;
    if (direction > 0)
        new_index = list.getNextWrap(list_index, mark);
    else
        new_index = list.getPreviousWrap(list_index, mark);

    if (new_index != 0 && new_index != list_index) {
        list_index = new_index;
        updateObject();
        changed = true;
    }
}

/** Toggle options. */
void
GChartLocation::toggleOptions(int32 which)
{
    getChartOpts(false, viewport.mult, viewport.divi).toggleOptions(which);
    changed = true;
}

bool
GChartLocation::isChanged() const
{
    return changed;
}

void
GChartLocation::markChanged()
{
    changed = true;
}

void
GChartLocation::clearChanged()
{
    clearWrapAdjust();
    changed = false;
}
