/**
  *  \file game/cargo-tr.cc
  *  \brief Cargo Transfer and Cargo Cost Transaction
  *
  *  This module implements the two most basic cargo transactions.
  *  GCargoTransfer is a means of moving cargo between objects.
  *  A GCargoCostTransaction is a transaction which turns cargo into
  *  something else.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"
#include "util/nls.h"

/***************************** GCargoTransfer ****************************/

/*! \class GCargoTransfer

    This class provides everything necessary to transfer cargo
    between two objects. */

/** Constructor.
    \param left left object
    \param right right object */
GCargoTransfer::GCargoTransfer(GCargoContainer& left, GCargoContainer& right)
{
    objs[0] = &left;
    objs[1] = &right;
}

/** Destructor. Cancels the transaction if it was not yet finished. */
GCargoTransfer::~GCargoTransfer()
{
    if (!isFinished())
        doCancel();
}

/** Move stuff.
    \param type    type to move
    \param amount  amount to move. May be negative; semantics are
                   preserved.
    \param from    source object (0=left, 1=right)
    \param partial true to allow partial transfers, false to reject
    \param s2m     sell supplies on-the-fly
    \return amount moved.
    \inv !partial => return == 0 || return == amount
    \inv partial  => return >= 0 && return <= amount */
int32
GCargoTransfer::move(GCargoType type, int32 amount, int from, bool partial, bool s2m)
{
    /* trivial case? */
    if (!amount)
        return 0;

    /* reverse transfer? */
    if (amount < 0)
        return -move(type, -amount, 1-from, partial, false);

    /* figure out src/dest objects and type */
    GCargoContainer& src  = *objs[from];
    GCargoContainer& dest = *objs[1 - from];

    GCargoType dtype = (type == el_Supplies && s2m ? el_Money : type);

    /* can we actually move it? */
    if (!src.canHaveCargo(type) || !dest.canHaveCargo(dtype))
        return 0;

    int32 space = dest.getMaxCargo(dtype) - dest.getCargo(dtype);
    if (space < 0)
        return 0;
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* do the transfer */
    int32 got = src.removeCargo(type, amount, partial);
    int32 put = dest.addCargo(dtype, got, partial);
    ASSERT(got == put);
    ASSERT(got <= amount);

    if (got)
        sig_changed.raise();

    return got;
}

/** Perform unload. Cargo containers have a notion of "unload source"
    and "unload target" (can_Unload, can_GetUnload). If this
    GCargoTransfer object has containers of both types, this function
    performs the unload operation by attempting to move all minerals,
    colonists, supplies and cash from the source to the target. Fuel
    and ammunition are not moved as we assume that the source unit (a
    ship) needs them for itself.

    \param s2m true to sell supplies on-the-fly

    \return true iff we had both a source and a target object (does
    not say whether any cargo was moved, just whether the structural
    requirement was fulfilled) */
bool
GCargoTransfer::doUnload(bool s2m)
{
    /* figure out direction */
    int direction;
    if (objs[0]->hasFlag(GCargoContainer::can_Unload) && objs[1]->hasFlag(GCargoContainer::can_GetUnload))
        direction = 0;
    else if (objs[1]->hasFlag(GCargoContainer::can_Unload) && objs[0]->hasFlag(GCargoContainer::can_GetUnload))
        direction = 1;
    else
        return false;

    /* transfer */
    move(el_Tritanium,  objs[direction]->getCargo(el_Tritanium),  direction, true, s2m);
    move(el_Duranium,   objs[direction]->getCargo(el_Duranium),   direction, true, s2m);
    move(el_Molybdenum, objs[direction]->getCargo(el_Molybdenum), direction, true, s2m);
    move(el_Colonists,  objs[direction]->getCargo(el_Colonists),  direction, true, s2m);
    move(el_Supplies,   objs[direction]->getCargo(el_Supplies),   direction, true, s2m);
    move(el_Money,      objs[direction]->getCargo(el_Money),      direction, true, s2m);
    return true;
}

void
GCargoTransfer::cancel()
{
    objs[0]->mustCancel();
    objs[1]->mustCancel();
}

void
GCargoTransfer::commit()
{
    objs[0]->mustCommit();
    objs[1]->mustCommit();
}
