/**
  *  \file int/if/conif.cc
  *  \brief Interpreter Game Interface - Console Components
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "int/compound.h"
#include "int/data.h"
#include "int/exec.h"
#include "int/if/conif.h"
#include "int/if/ifutil.h"
#include "int/intglobal.h"
#include "int/names.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "int/args.h"

/** Global command "UI.Input", console version. */
static void
IFUIInput(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Input <prompt>[, <title>, <maxChars>, <flags>, <default>] */
    args.checkArgumentCount(1, 5);

    string_t prompt;
    string_t title;
    string_t defaultText;
    int32_t  maxChars = 255;
    int32_t  flags = 0;
    int32_t  width = 0;

    // Mandatory argument
    if (!checkStringArg(prompt, args.getNext()))
        return;

    // Optional arguments
    title = prompt;
    checkStringArg(title, args.getNext());
    checkIntArg(maxChars, args.getNext(), 0, 32000);
    checkFlagArg(flags, &width, args.getNext(), "NHPFGM");
    checkStringArg(defaultText, args.getNext());

    // Convert flags
    //   N = 1 = numeric
    //   H = 2 = on high ASCII
    //   P = 4 = password masking
    //   F = 8 = frame
    //   G = 16 = game charset
    //   M = 32 = width is in ems
    while (1) {
        // Show prompt and get initial input
        string_t line;
        std::cout << prompt << "> " << std::flush;
        if (!getline(std::cin, line)) {
            // EOF
            exc.setVariable("UI.RESULT", 0);
            return;
        }

        if (line.size() > uint32_t(maxChars))
            line.erase(maxChars);

        // Check whether this is valid
        bool hadInvalidChars = false;
        for (string_t::size_type i = 0; i < line.size(); ++i) {
            uint8_t c = line[i];
            if (((c >= '0' && c <= '9')
                 || (c >= ' ' && c <= 126 && (flags & 1) == 0)
                 || (c >= 128 && c <= 255 && (flags & 3) == 0))
                && ((flags & 16) == 0 || Recoder::isReliableChar(c)))
            {
                // Character is safe (this is the same condition that is used in UIInputLine::handleEvent
            } else {
                // Character is not accepted
                hadInvalidChars = true;
                line[i] = (flags & 1) ? '0' : '?';
            }
        }
        if (!hadInvalidChars) {
            // Accepted input
            IntStringValue sv(line);
            exc.setVariable("UI.RESULT", &sv);
            return;
        }

        // Not accepted, continue
        std::cout << _("Your input contains characters that are not permitted at this place.\nPlease try again.\n");
    }
}


static void
initInterpreterConsoleInterface()
{
    defineGlobalValue("SYSTEM.GUI", makeBoolValue(0));
    defineGlobalProcedure("UI.INPUT", IFUIInput);
    // MessageBox -> core.q
    // UI.Update -> core.q
}

// Constraints: after mod_intgame_common, after mod_int, with mod_base_console
const ModuleInitializer mod_intgame_console[] = {
    { OnStartup, initInterpreterConsoleInterface },
    { 0, 0 }
};
