/**
  *  \file int/statement.cc
  *  \brief Interpreter: Compiling statements
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include <cstring>
#include "int/statement.h"
#include "int/interror.h"
#include "int/parseexpr.h"
#include "int/keywords.h"
#include "int/unary.h"
#include "int/context.h"
#include "int/exec.h"
#include "int/compound.h"
#include "int/binary.h"
#include "int/ternary.h"
#include "int/builtin.h"
#include "int/args.h"



static string_t
stripPrefix(string_t s, const char* pfx)
{
    std::size_t pfxlen = std::strlen(pfx);
    if (s.compare(0, pfxlen, pfx, pfxlen) == 0) {
        if (s.size() == pfxlen)
            throw IntError("Invalid identifier");
        s.erase(0, pfxlen);
    }
    return s;
}

/********************* IntStatementCompilationContext ********************/

/** Constructor. */
IntStatementCompilationContext::IntStatementCompilationContext()
    : IntCompilationContext(),
      parent(0),
      exec_context(0)
{ }

/** Construct child context.
    \param parent Parent. We inherit all properties from it. */
IntStatementCompilationContext::IntStatementCompilationContext(const IntStatementCompilationContext* parent)
    : IntCompilationContext(*parent),
      parent(parent),
      exec_context(parent->exec_context)
{ }

/** Generate code for "Break" statement.
    The default implementation delegates this to the parent, or fails if we don't have one. */
void
IntStatementCompilationContext::compileBreak(IntBytecodeObject& bco) const
{
    if (parent)
        parent->compileBreak(bco);
    else
        throw IntError::misplacedKeyword("Break");
}

/** Generate code for "Continue" statement.
    The default implementation delegates this to the parent, or fails if we don't have one. */
void
IntStatementCompilationContext::compileContinue(IntBytecodeObject& bco) const
{
    if (parent)
        parent->compileContinue(bco);
    else
        throw IntError::misplacedKeyword("Continue");
}

/** Generate code to clean up the stack for a "Return" statement.
    The default implementation delegates this to the parent, or does nothing if we don't have one.
    Since exiting a frame cleans up its exception and context, we only need to clean up stack
    temporaries here. */
void
IntStatementCompilationContext::compileCleanup(IntBytecodeObject& bco) const
{
    if (parent)
        parent->compileCleanup(bco);
}


/************************** IntStatementCompiler *************************/

/** Constructor.
    \param cs Command source that contains statements to compile. */
IntStatementCompiler::IntStatementCompiler(IntCommandSource& cs)
    : cs(cs)
{
    cs.readNextLine();
}

/** Compile a statement. Assumes input at beginning of statement, that
    is, the tokenizer initialized to the current line, looking at the
    first token. Leaves input at end of statement. That is, caller
    must invoke cs.readNextLine() to advance to the next line. On a
    multi-line statement, this leaves the input after the end of the
    terminator.
    \param bco [out] Bytecode object that contains code
    \param scc [in] Statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compile(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* End of file reached? */
    if (cs.isEOF()) {
        if (scc.hasFlags(IntCompilationContext::WantTerminators)) {
            /* If we want a terminator, we do not want end of file */
            throw IntError("Unexpected end of script");
        } else {
            return EndOfInput;
        }
    }

    /* Remember current compilation position for debugging */
    bco.addLineNumber(cs.getLineNumber());

    /* Do we have a nice statement? */
    IntTokenizer& tok = cs.getTokenizer();
    if (tok.getCurrentToken() == tok.tEnd) {
        /* Blank line */
        return CompiledStatement;
    } else if (tok.getCurrentToken() == tok.tIdentifier) {
        /* Identifier */
        switch (lookupKeyword(tok.getCurrentString())) {
         case kwAbort:
            return compileAbort(bco, scc);

         case kwBind:
            return compileBind(bco, scc);

         case kwBreak:
            tok.readNext();
            parseEndOfLine();
            scc.compileBreak(bco);
            return CompiledStatement;

            // case kwClose: -- files, beta 10; regular subroutine?

         case kwContinue:
            tok.readNext();
            parseEndOfLine();
            scc.compileContinue(bco);
            return CompiledStatement;

         case kwCreateKeymap:
            return compileCreateKeymap(bco, scc);
            
         case kwCreatePlanetProperty:
            return compileCreateProperty(bco, scc, IntOpcode::miSpecialDefPlanetProperty, "PLANET.");

         case kwCreateShipProperty:
            return compileCreateProperty(bco, scc, IntOpcode::miSpecialDefShipProperty, "SHIP.");

         case kwDim:
            tok.readNext();
            return compileDim(bco, scc, IntOpcode::sLocal, 0);
         case kwDo:
            return compileDo(bco, scc);

         // case kwEnd: -- processes, beta 9 or 10

         case kwCase:
         case kwElse:
         case kwEndIf:
         case kwEndSelect:
         case kwEndSub:
         case kwEndFunction:
         case kwEndTry:
         case kwEndWith:
         case kwLoop:
         case kwNext:
            if (scc.hasFlags(IntCompilationContext::WantTerminators))
                return Terminator;
            else
                throw IntError::misplacedKeyword(tok.getCurrentString().c_str());
         case kwEval:
            return compileEval(bco, scc);
         case kwFor:
            return compileFor(bco, scc);
         case kwForEach:
            return compileForEach(bco, scc);
         case kwFunction:
            return compileSub(bco, scc, false);

         // case kwGet: -- files, beta 10

         case kwIf:
            return compileIf(bco, scc);

         // case kwInput: -- files, beta 10

         case kwLoad:
            return compileLoad(bco, scc, true);
         case kwLocal:
            tok.readNext();
            return compileDim(bco, scc, IntOpcode::sLocal, 1);

         case kwOn:
            return compileOn(bco, scc);

         // case kwOpen: -- files, beta 10
         // case kwOption: -- optimisation. when?

         case kwPrint:
            return compilePrint(bco, scc);

         // case kwPut: -- files, beta 10; regular subroutine?

         case kwRestart:
            throw IntError::misplacedKeyword("Restart");

         case kwReturn:
            tok.readNext();
            return compileReturn(bco, scc);

         case kwRunHook:
            return compileRunHook(bco, scc);

         // case kwSeek: -- files, beta 10; regular subroutine?

         case kwSelect:
            return compileSelect(bco, scc);

         // case kwSelectionExec: -- multi-selection. when?
         // case kwSetByte: -- files; beta 10
         // case kwSetLong: -- files; beta 10
         // case kwSetStr: -- files; beta 10
         // case kwSetWord: -- files; beta 10

         case kwShared:
            tok.readNext();
            return compileDim(bco, scc, IntOpcode::sShared, 1);
         case kwStatic:
            tok.readNext();
            return compileDim(bco, scc, IntOpcode::sStatic, 1);

         // case kwStop: -- processes, beta 10

         case kwSub:
            return compileSub(bco, scc, true);
         case kwTry:
            return compileTry(bco, scc);
         case kwTryLoad:
            return compileLoad(bco, scc, false);
         case kwUntil:
            throw IntError::misplacedKeyword("Until");
         case kwWhile:
            throw IntError::misplacedKeyword("While");
         case kwWith:
            return compileWith(bco, scc);
         case kwNone:
            return compileAmbiguousStatement(bco, scc);

         default:
            throw IntError("Not implemented");
        }
    } else {
        /* Anything else, must be expression */
        return compileExpressionStatement(bco, scc);
    }
}

/** Compile statement list.
    \param bco [out] Bytecode object that contains code
    \param scc [in] Statement compilation context
    \retval EndOfInput We ended the compilation because the input ended (only if !scc.hasFlags(WantTerminators))
    \retval Terminator We ended the compilation because we found a terminator (only if scc.hasFlags(WantTerminators)) */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileList(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    while (1) {
        switch (StatementResult r = compile(bco, scc)) {
         case EndOfInput:
         case Terminator:
            return r;
         case CompiledExpression:
            /* This means there is a value on top of the stack which we do not want */
            bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
            cs.readNextLine();
            break;
         default:
            cs.readNextLine();
            break;
        }
    }
}

/** Compile ambiguous statement. If a statement starts with an unknown keyword, it might
    be an expression or a subroutine call. Both have distinct syntax, and even some
    identical productions: "SetSpeed +3" could be a subroutine call using argument "+3",
    or an expression performing an addition and discarding the result.

    If we have an execution context, we're lucky and can just look up what we find; it's
    an error if we find nothing.

    If we don't have an execution context, we'll just generate an "Eval" instruction.
    To speed up execution, we apply a quick check, though: we check the token following the
    keyword. It can be either a possible first token in an expression, or a possible second
    one. If it is a possible first, but not a possible second, we have a subroutine call.
    If it is a possible second, but not a possible first, we have an expression.

    Possible seconds are all binary operators, parentheses ("foo(x)") and dot.
    Possible firsts are all unary operators, identifiers, literals, parentheses */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileAmbiguousStatement(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    if (IntExecutionContext* exc = scc.getExecutionContext()) {
        /* We have an execution context, so we can actually look up the value to see what it is. */
        bool is_proc;
        int32_t index;
        if (IntContext* con = exc->lookup(cs.getTokenizer().getCurrentString(), index)) {
            std::auto_ptr<IntValue> v(con->get(index));
            if (IntCallableValue* cv = dynamic_cast<IntCallableValue*>(v.get())) {
                // Callable builtin
                is_proc = cv->isProcedureCall();
            } else if (IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(v.get())) {
                // Callable subroutine
                is_proc = sv->getBytecodeObject()->isProcedure();
            } else {
                // Variable
                is_proc = false;
            }
        } else {
            // Unknown; expression may be valid if this is a builtin fundamental function
            if (!lookupBuiltinFunction(cs.getTokenizer().getCurrentString()))
                throw IntError::unknownIdentifier(cs.getTokenizer().getCurrentString());
            is_proc = false;
        }

        if (is_proc) {
            /* It's a procedure */
            return compileProcedureCall(bco, scc);
        } else {
            /* It's a function */
            return compileExpressionStatement(bco, scc);
        }
    } else {
        /* Find next token */
        IntTokenizer::TokenType next = IntTokenizer(cs.getTokenizer()).readNext();
        switch (next) {
         case IntTokenizer::tInteger:
         case IntTokenizer::tFloat:
         case IntTokenizer::tString:
         case IntTokenizer::tBoolean:
         case IntTokenizer::tIdentifier:
         case IntTokenizer::tNOT:
            /* Possible firsts but not possible seconds: subroutine call */
            return compileProcedureCall(bco, scc);

         case IntTokenizer::tAND:
         case IntTokenizer::tOR:
         case IntTokenizer::tXOR:
         case IntTokenizer::tMOD:
         case IntTokenizer::tNE:
         case IntTokenizer::tGE:
         case IntTokenizer::tLE:
         case IntTokenizer::tAssign:
         case IntTokenizer::tAmpersand:
         case IntTokenizer::tMultiply:
         case IntTokenizer::tSlash:
         case IntTokenizer::tBackslash:
         case IntTokenizer::tCaret:
         case IntTokenizer::tEQ:
         case IntTokenizer::tLT:
         case IntTokenizer::tGT:
         case IntTokenizer::tDot:
         case IntTokenizer::tSemicolon:
            /* Possible seconds but not possible firsts: expression */
            return compileExpressionStatement(bco, scc);

         case IntTokenizer::tComma:
         case IntTokenizer::tInvalid:
         case IntTokenizer::tRParen:
         case IntTokenizer::tColon:
            /* Impossible anywhere */
            throw IntError("Syntax error");

         case IntTokenizer::tHash:
         case IntTokenizer::tPlus:
         case IntTokenizer::tMinus:
         case IntTokenizer::tLParen:
         case IntTokenizer::tEnd:
         default:
            /* Possible firsts and seconds: defer until execution */
            IntStringValue sv(cs.getTokenizer().getCurrentString() + cs.getTokenizer().getRemainingLine());
            bco.addPushLiteral(&sv);
            bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEvalStatement, 1);
            return CompiledStatement;
        }
    }
}

/** Compile "Abort" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileAbort(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Parse args */
    ptr_vector<IntExprNode> args;
    cs.getTokenizer().readNext();
    parseArgumentList(args);
    checkArgumentCount(args.size(), 0, 1);

    /* Compile */
    if (args.size() == 0)
        bco.addPushLiteral(0);
    else
        args[0]->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maUnary, unThrow, 0);
    return CompiledStatement;
}

/** Compile "Bind" statement.
    \param bco Bytecode output
    \param scc Statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileBind(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Bind keymap <expr> := <expr> [, <expr> := <expr>]*
           pushlit "keymap"
           ukeylookup
          [<expr>
           <expr>
           tkeyadd]*
           drop 1 */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    if (tok.getCurrentToken() != tok.tIdentifier)
        throw IntError::expectIdentifier("keymap name");

    /* Push keymap */
    IntStringValue keymap(tok.getCurrentString());
    bco.addPushLiteral(&keymap);
    bco.addInstruction(IntOpcode::maUnary, unKeyLookup, 0);
    tok.readNext();

    /* Parse assignments */
    while (1) {
        std::auto_ptr<IntExprNode> expr(IntExpressionParser(cs.getTokenizer()).parseNA());
        expr->compileValue(bco, scc);

        /* Here, we only accept ":=", because "=" is swallowed by parseNA anyway. */
        if (!tok.checkAdvance(tok.tAssign))
            throw IntError::expectSymbol(":=");

        expr.reset(IntExpressionParser(cs.getTokenizer()).parseNA());
        expr->compileValue(bco, scc);

        bco.addInstruction(IntOpcode::maTernary, teKeyAdd, 0);

        if (tok.getCurrentToken() == tok.tEnd) {
            break;
        } else if (tok.checkAdvance(tok.tComma)) {
            /* ok */
        } else {
            throw IntError::expectSymbol(",");
        }
    }

    /* Drop keymap */
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
    return CompiledStatement;
}

/** Compile "CreateKeymap" statement.
    \param bco Bytecode output
    \param scc Statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileCreateKeymap(IntBytecodeObject& bco, const IntStatementCompilationContext& /*scc*/)
{
    /* CreateKeymap keymap(parent...), ...
           pushlit "keymap"
           ukeycreate
          [pushlit "parent
           ukeylookup
           bkeyaddparent]
           drop 1 */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    while (1) {
        if (tok.getCurrentToken() != tok.tIdentifier)
            throw IntError::expectIdentifier("keymap name");

        IntStringValue keymapName(tok.getCurrentString());
        bco.addPushLiteral(&keymapName);
        bco.addInstruction(IntOpcode::maUnary, unKeyCreate, 0);
        tok.readNext();
        if (tok.checkAdvance(tok.tLParen)) {
            if (!tok.checkAdvance(tok.tRParen)) {
                while (1) {
                    if (tok.getCurrentToken() != tok.tIdentifier)
                        throw IntError::expectIdentifier("parent keymap name");
                    IntStringValue parentName(tok.getCurrentString());
                    bco.addPushLiteral(&parentName);
                    bco.addInstruction(IntOpcode::maUnary, unKeyLookup, 0);
                    bco.addInstruction(IntOpcode::maBinary, biKeyAddParent, 0);
                    tok.readNext();
                    if (tok.checkAdvance(tok.tComma)) {
                        // continue
                    } else if (tok.checkAdvance(tok.tRParen)) {
                        // end
                        break;
                    } else {
                        throw IntError::expectSymbol(",", ")");
                    }
                }
            }
        }
        bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);

        if (tok.checkAdvance(tok.tComma)) {
            // continue
        } else if (tok.getCurrentToken() == tok.tEnd) {
            // end
            break;
        } else {
            throw IntError::garbageAtEnd(false);
        }
    }
    return CompiledStatement;
}

/** Compile "CreateShipProperty" or "CreatePlanetProperty" statement.
    \param bco Bytecode output
    \param scc Statement compilation context
    \param scope Scope given for variables
    \param num_scopes Number of scopes already given in statement */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileCreateProperty(IntBytecodeObject& bco, const IntStatementCompilationContext& /*scc*/, IntOpcode::Special minor, const char* prefix)
{
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    while (1) {
        if (tok.getCurrentToken() != tok.tIdentifier)
            throw IntError::expectIdentifier("property name");
        bco.addInstruction(IntOpcode::maSpecial, minor, bco.addName(stripPrefix(tok.getCurrentString(), prefix)));
        tok.readNext();
        if (tok.checkAdvance(tok.tComma)) {
            // OK, another one
        } else if (tok.getCurrentToken() == tok.tEnd) {
            // OK, end
            break;
        } else {
            // Error
            throw IntError::garbageAtEnd(false);
        }
    }
    return CompiledStatement;
}

/** Compile "Dim" statement.
    \param bco Bytecode output
    \param scc Statement compilation context
    \param scope Scope given for variables
    \param num_scopes Number of scopes already given in statement */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileDim(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntOpcode::Scope scope, int num_scopes)
{
    /* Read scope */
    IntTokenizer& tok = cs.getTokenizer();
    while (1) {
        if (tok.checkAdvance("LOCAL")) {
            scope = IntOpcode::sLocal;
            if (num_scopes > 0)
                throw IntError::misplacedKeyword("Local");
            ++num_scopes;
        } else if (tok.checkAdvance("STATIC")) {
            scope = IntOpcode::sStatic;
            if (num_scopes > 0)
                throw IntError::misplacedKeyword("Static");
            ++num_scopes;
        } else if (tok.checkAdvance("SHARED")) {
            scope = IntOpcode::sShared;
            if (num_scopes > 0)
                throw IntError::misplacedKeyword("Shared");
            ++num_scopes;
        } else {
            break;
        }
    }

    /* Read variables */
    while (1) {
        /* Read name */
        if (tok.getCurrentToken() != tok.tIdentifier)
            throw IntError::expectIdentifier("variable name");
        string_t name = tok.getCurrentString();
        if (scope == IntOpcode::sShared)
            name = stripPrefix(name, "GLOBAL.");
        validateName(name);
        tok.readNext();

        /* Read value */
        bool isNull;
        if (tok.checkAdvance(tok.tEQ) || tok.checkAdvance(tok.tAssign)) {
            compileArgumentExpression(bco, scc);
            isNull = false;
        } else {
            isNull = true;
        }

        /* Optimisation: if this is going to be a local variable, and we are linearly
           executing in local context, we can pre-allocate this name, to allow future
           references to use the address instead of a costly name lookup. We only do
           this if the variable hasn't been mentioned yet, to avoid retroactively turning
           global references into locals. This is overly pessimistic because it prevents
           things like 'Local name = Ship(Id).Name' from being optimized ('name' has
           already been mentioned by the above compileExpression before this check,
           as a member name), but it should be safe.

           Note that a variable can also be used by code called from this subroutine.
           This is checked by hasUserCall(). */
        if (scope == IntOpcode::sLocal
            && scc.hasFlags(IntCompilationContext::LinearExecution)
            && scc.hasFlags(IntCompilationContext::LocalContext)
            && !bco.hasName(name)
            && !bco.hasUserCall())
        {
            /* Optimized version */
            if (bco.hasLocalVariable(name)) {
                /* We know that this is a duplicate, so throw away the value */
                // FIXME: generate warnings
                if (!isNull)
                    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
            } else {
                /* We know that this is a new variable, so initialize it */
                bco.addLocalVariable(name);
                if (!isNull)
                    bco.addVariableReferenceInstruction(IntOpcode::maPop, name, scc);
            }
        } else {
            /* General version */
            if (isNull)
                bco.addPushLiteral(0);
            bco.addInstruction(IntOpcode::maDim, scope, bco.addName(name));
        }

        if (!tok.checkAdvance(tok.tComma))
            break;
    }

    if (tok.getCurrentToken() != tok.tEnd)
        throw IntError::expectSymbol(",");

    return CompiledStatement;
}

/** Compile "Do" (generic loop) statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileDo(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Do [While <a1>|Until <a2>] / <body> / Loop [While <e1>|Until <e2>]

       again:
         <a1/a2>
       do:
         <body>
       continue:
         <e1/e2>/j again
       break: */

    /* Context for child code */
    struct DoStatementCompilationContext : public IntStatementCompilationContext {
        DoStatementCompilationContext(const IntStatementCompilationContext* parent,
                                      IntBytecodeObject::label_t lcontinue,
                                      IntBytecodeObject::label_t lbreak)
            : IntStatementCompilationContext(parent),
              lcontinue(lcontinue),
              lbreak(lbreak)
            {
                withoutFlags(LinearExecution);
                setBlockSyntax();
            }
        void compileBreak(IntBytecodeObject& bco) const
            {
                bco.addJump(IntOpcode::jAlways, lbreak);
            }
        void compileContinue(IntBytecodeObject& bco) const
            {
                bco.addJump(IntOpcode::jAlways, lcontinue);
            }

        IntBytecodeObject::label_t lcontinue, lbreak;
    };

    /* Allowed? */
    if (scc.hasFlags(IntCompilationContext::RefuseBlocks))
        throw IntError::invalidMultiline();

    /* Make labels */
    IntBytecodeObject::label_t lagain    = bco.makeLabel();
    IntBytecodeObject::label_t ldo       = bco.makeLabel();
    IntBytecodeObject::label_t lbreak    = bco.makeLabel();
    IntBytecodeObject::label_t lcontinue = bco.makeLabel();
    bco.addLabel(lagain);

    /* Compile head condition */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    if (tok.checkAdvance("WHILE")) {
        compileArgumentCondition(bco, scc, ldo, lbreak);
    } else if (tok.checkAdvance("UNTIL")) {
        compileArgumentCondition(bco, scc, lbreak, ldo);
    }
    parseEndOfLine();

    /* Compile body */
    cs.readNextLine();
    bco.addLabel(ldo);
    compileList(bco, DoStatementCompilationContext(&scc, lcontinue, lbreak));
    if (!tok.checkAdvance("LOOP"))
        throw IntError::expectKeyword("Loop");

    /* Compile tail condition */
    bco.addLabel(lcontinue);
    if (tok.checkAdvance("UNTIL")) {
        compileArgumentCondition(bco, scc, lbreak, lagain);
    } else if (tok.checkAdvance("WHILE")) {
        compileArgumentCondition(bco, scc, lagain, lbreak);
    } else {
        bco.addJump(IntOpcode::jAlways, lagain);
    }
    parseEndOfLine();

    bco.addLabel(lbreak);
    return CompiledBlock;
}

/** Compile "Eval" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileEval(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Skip over "Eval" token */
    cs.getTokenizer().readNext();

    /* Read arguments */
    ptr_vector<IntExprNode> args;
    parseArgumentList(args);
    if (args.size() == 0)
        throw IntError("Too few arguments to 'Eval'");

    /* Compile */
    for (ptr_vector<IntExprNode>::size_type i = 0; i != args.size(); ++i)
        args[i]->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEvalStatement, args.size());

    return CompiledStatement;
}

/** Compile "For" (counting loop) statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileFor(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* For <var> := <start> To <end> Do <body> */

    /* Context for child block */
    struct ForStatementCompilationContext : public IntStatementCompilationContext {
        ForStatementCompilationContext(const IntStatementCompilationContext* parent,
                                       IntBytecodeObject::label_t lcontinue,
                                       IntBytecodeObject::label_t lbreak)
            : IntStatementCompilationContext(parent),
              lcontinue(lcontinue),
              lbreak(lbreak)
            {
                withoutFlags(LinearExecution);
            }
        void compileBreak(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
                bco.addJump(IntOpcode::jAlways, lbreak);
            }
        void compileContinue(IntBytecodeObject& bco) const
            {
                bco.addJump(IntOpcode::jAlways, lcontinue);
            }
        void compileCleanup(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
                IntStatementCompilationContext::compileCleanup(bco);
            }

        IntBytecodeObject::label_t lcontinue, lbreak;
    };

    /* Parse it */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();

    /* ...induction variable... */
    if (tok.getCurrentToken() != tok.tIdentifier)
        throw IntError::expectIdentifier("variable name");
    const string_t var = tok.getCurrentString();
    tok.readNext();

    if (!tok.checkAdvance(tok.tEQ) && !tok.checkAdvance(tok.tAssign))
        throw IntError::expectSymbol("=", ":=");

    /* ...start expression... */
    std::auto_ptr<IntExprNode> start(IntExpressionParser(tok).parse());
    if (!tok.checkAdvance("TO"))
        throw IntError::expectKeyword("To");

    /* ...end expression... */
    std::auto_ptr<IntExprNode> end(IntExpressionParser(tok).parse());

    bool oneliner = tok.checkAdvance("DO");

    /* Generate code for head */
    /*
                  <end>                 ; b             ; break must include drop
                 [upos]                                 ; ensure it is a number
                  <start>               ; b:a
                 [upos]                                 ; ensure it is a number
           again: store <var>           ; b:a
                  dup 1                 ; b:a:b
                  bcmple                ; b:res
                  jfep out              ; b
                  <body>
        continue: push <var>            ; b:i
                  uinc                  ; b:1+i
                  j again
             out: drop 1                ;
           break: */
    IntBytecodeObject::label_t lagain    = bco.makeLabel();
    IntBytecodeObject::label_t lcontinue = bco.makeLabel();
    IntBytecodeObject::label_t lout      = bco.makeLabel();
    IntBytecodeObject::label_t lbreak    = bco.makeLabel();

    end->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maUnary, unPos, 0);
    start->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maUnary, unPos, 0);
    bco.addLabel(lagain);
    bco.addVariableReferenceInstruction(IntOpcode::maStore, var, scc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDup, 1);
    bco.addInstruction(IntOpcode::maBinary, biCompareLE, 0);          /* no need to handle CaseBlind, we're dealing with numbers */
    bco.addJump(IntOpcode::jIfFalse | IntOpcode::jIfEmpty | IntOpcode::jPopAlways, lout);

    /* Body */
    StatementResult result;
    if (tok.getCurrentToken() == tok.tEnd) {
        /* Multi-line */
        if (scc.hasFlags(scc.RefuseBlocks))
            throw IntError::invalidMultiline();
        cs.readNextLine();
        compileList(bco, ForStatementCompilationContext(&scc, lcontinue, lbreak).setBlockSyntax());
        if (!tok.checkAdvance("NEXT"))
            throw IntError::expectKeyword("Next");
        parseEndOfLine();
        result = CompiledBlock;
    } else {
        /* One-liner */
        if (!oneliner)
            throw IntError::expectKeyword("Do");
        compile(bco, ForStatementCompilationContext(&scc, lcontinue, lbreak).setOneLineSyntax());
        result = CompiledStatement;
    }

    /* Compile tail */
    bco.addLabel(lcontinue);
    bco.addVariableReferenceInstruction(IntOpcode::maPush, var, scc);
    bco.addInstruction(IntOpcode::maUnary, unInc, 0);
    bco.addJump(IntOpcode::jAlways, lagain);
    bco.addLabel(lout);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
    bco.addLabel(lbreak);

    return result;
}

/** Compile "ForEach" (set loop) statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileForEach(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /*  <expr>         break: endindex
        sfirstindex           j end
        jfep end
     again:
        <body>
     continue:
        snextindex
        jtp again
     end: */

    IntTokenizer& tok = cs.getTokenizer();

    /* Compile scope expression */
    tok.readNext();
    std::auto_ptr<IntExprNode> scope_expr(IntExpressionParser(tok).parse());

    /* Make label */
    IntBytecodeObject::label_t lagain    = bco.makeLabel();
    IntBytecodeObject::label_t lend      = bco.makeLabel();
    IntBytecodeObject::label_t lcontinue = bco.makeLabel();

    /* Context for child code */
    struct ForEachStatementCompilationContext : public IntStatementCompilationContext {
        ForEachStatementCompilationContext(const IntStatementCompilationContext* parent,
                                           IntBytecodeObject::label_t lcontinue,
                                           IntBytecodeObject::label_t lend)
            : IntStatementCompilationContext(parent),
              lcontinue(lcontinue),
              lend(lend)
            {
                withoutFlags(LocalContext | LinearExecution);
                withExecutionContext(0);
            }
        void compileBreak(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndIndex, 0);
                bco.addJump(IntOpcode::jAlways, lend);
            }
        void compileContinue(IntBytecodeObject& bco) const
            {
                bco.addJump(IntOpcode::jAlways, lcontinue);
            }

        IntBytecodeObject::label_t lcontinue, lend;
    };

    ForEachStatementCompilationContext subcc(&scc, lcontinue, lend);

    /* Compile loop head */
    scope_expr->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialFirstIndex, 0);
    bco.addJump(IntOpcode::jIfFalse | IntOpcode::jIfEmpty | IntOpcode::jPopAlways, lend);
    bco.addLabel(lagain);

    /* Compile loop body */
    bool oneliner = tok.checkAdvance("DO");
    StatementResult result;
    if (tok.getCurrentToken() != tok.tEnd) {
        /* Single line */
        if (!oneliner)
            throw IntError::expectKeyword("Do");
        subcc.setOneLineSyntax();
        compile(bco, subcc);
        result = CompiledStatement;
    } else {
        /* Multi-line */
        if (scc.hasFlags(IntCompilationContext::RefuseBlocks))
            throw IntError::invalidMultiline();

        subcc.setBlockSyntax();
        cs.readNextLine();
        compileList(bco, subcc);
        if (!tok.checkAdvance("NEXT"))
            throw IntError::expectKeyword("Next");
        parseEndOfLine();
        result = CompiledBlock;
    }

    /* Compile loop tail */
    bco.addLabel(subcc.lcontinue);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialNextIndex, 0);
    bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, lagain);
    bco.addLabel(lend);

    return result;
}

/** Compile "If" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileIf(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    IntTokenizer& tok = cs.getTokenizer();

    /* Make labels */
    IntBytecodeObject::label_t ift = bco.makeLabel();
    IntBytecodeObject::label_t iff = bco.makeLabel();

    /* Read expression */
    tok.readNext();
    compileArgumentCondition(bco, scc, ift, iff);
    bco.addLabel(ift);

    /* Does this look like a one-liner? */
    bool oneliner = tok.checkAdvance("THEN");

    /* Single or multiple lines? */
    if (tok.getCurrentToken() != tok.tEnd) {
        /* Single line */
        if (!oneliner)
            throw IntError::expectKeyword("Then");

        /* Compile 'Then' (will return CompiledStatement) */
        compile(bco, IntStatementCompilationContext(&scc).setOneLineSyntax().withoutFlags(IntCompilationContext::LinearExecution));
        bco.addLabel(iff);

        /* Process result */
        return CompiledStatement;
    } else {
        /* Multiple lines */
        if (scc.hasFlags(IntCompilationContext::RefuseBlocks))
            throw IntError::invalidMultiline();
        cs.readNextLine();

        /* Compile 'Then' part. */
        compileList(bco, IntStatementCompilationContext(&scc).setBlockSyntax().withoutFlags(IntCompilationContext::LinearExecution));

        /* Compile 'Else' and 'Else If' parts. hadElse is true when we had an Else part,
           and do not allow another. 'iff' will always be placed at the Else part. */
        IntBytecodeObject::label_t endif = bco.makeLabel();
        bool hadElse = false;
        while (1) {
            if (tok.checkAdvance("ELSE")) {
                bco.addJump(IntOpcode::jAlways, endif);
                bco.addLabel(iff);
                if (tok.checkAdvance("IF")) {
                    /* If / ... / Else If expr / ... */
                    ift = bco.makeLabel();
                    iff = bco.makeLabel();
                    compileArgumentCondition(bco, scc, ift, iff);
                    bco.addLabel(ift);
                    tok.checkAdvance("THEN");
                } else {
                    if (hadElse)
                        throw IntError::misplacedKeyword("Else");
                    hadElse = true;
                }
                parseEndOfLine();
                compileList(bco, IntStatementCompilationContext(&scc).setBlockSyntax().withoutFlags(IntCompilationContext::LinearExecution));
            } else if (tok.checkAdvance("ENDIF")) {
                parseEndOfLine();
                bco.addLabel(endif);
                if (!hadElse)
                    bco.addLabel(iff);
                break;
            } else {
                throw IntError::expectKeyword("Else", "EndIf");
            }
        }
        return CompiledBlock;
    }
}

/** Compile "Load" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileLoad(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, bool mustSucceed)
{
    /*      <expr>
            sload
           [je 1F
            uthrow
       1H:] drop 1 */

    /* Parse it */
    cs.getTokenizer().readNext();
    compileArgumentExpression(bco, scc);
    parseEndOfLine();

    /* Generate code */
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialLoad, 0);
    if (mustSucceed) {
        IntBytecodeObject::label_t lab = bco.makeLabel();
        bco.addJump(IntOpcode::jIfEmpty, lab);
        bco.addInstruction(IntOpcode::maUnary, unThrow, 0);
        bco.addLabel(lab);
    }
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);

    return CompiledStatement;
}

/** Compile "On" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileOn(IntBytecodeObject& bco, const IntStatementCompilationContext& /*scc*/)
{
    /* Parse */
    IntTokenizer& tok = cs.getTokenizer();
    if (tok.readNext() != tok.tIdentifier)
        throw IntError::expectIdentifier("hook name");
    string_t name = tok.getCurrentString();
    tok.readNext();
    if (!tok.checkAdvance("DO"))
        throw IntError::expectKeyword("Do");
    if (tok.getCurrentToken() == tok.tEnd)
        throw IntError::invalidMultiline(); // FIXME: is this a good error message?

    /* Context for embedded command: one-liner, behaves like regular
       function body, Return not allowed. */
    class HookCompilationContext : public IntStatementCompilationContext {
     public:
        HookCompilationContext()
            {
                setOneLineSyntax();
                withFlags(LocalContext | LinearExecution);
            }
        void compileCleanup(IntBytecodeObject& /*bco*/) const
            { throw IntError::misplacedKeyword("Return"); }
    };

    /* Compile embedded command */
    IntBCORef nbco = new IntBytecodeObject();
    nbco->setIsProcedure(true);
    nbco->setFileName(bco.getFileName());
    compile(*nbco, HookCompilationContext());

    /* Compile this command */
    IntStringValue sv(name);
    bco.addPushLiteral(&sv);
    bco.addInstruction(IntOpcode::maPush,    IntOpcode::sSubroutine,      bco.addSubroutine(nbco));
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialAddHook, 0);
    nbco->relocate();

    return CompiledStatement;
}

/** Compile "Print" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compilePrint(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Parse it */
    ptr_vector<IntExprNode> nodes;
    cs.getTokenizer().readNext();
    parseArgumentList(nodes);

    /* Compile it */
    if (nodes.size() == 0) {
        IntStringValue sv("");
        bco.addPushLiteral(&sv);
    } else {
        nodes[0]->compileValue(bco, scc);
        for (ptr_vector<IntExprNode>::size_type i = 1; i < nodes.size(); ++i) {
            nodes[i]->compileValue(bco, scc);
            bco.addInstruction(IntOpcode::maBinary, biConcatEmpty, 0);
        }
    }
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialPrint, 0);
    return CompiledStatement;
}

/** Compile "Return" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileReturn(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Prepare */
    scc.compileCleanup(bco);

    /* Compile instruction */
    if (!bco.isProcedure()) {
        compileArgumentExpression(bco, scc);
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialReturn, 1);
    } else {
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialReturn, 0);
    }

    /* Must now be at end */
    parseEndOfLine();

    return CompiledStatement;
}

/** Compile "RunHook" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileRunHook(IntBytecodeObject& bco, const IntStatementCompilationContext& /*scc*/)
{
    /* Fetch name */
    IntTokenizer& tok = cs.getTokenizer();
    if (tok.readNext() != tok.tIdentifier)
        throw IntError::expectIdentifier("hook name");

    /* Compile */
    IntStringValue sv(tok.getCurrentString());
    bco.addPushLiteral(&sv);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialRunHook, 0);

    /* Most now be at end */
    tok.readNext();
    parseEndOfLine();

    return CompiledStatement;
}

/** Compile "Select Case" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileSelect(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* The selector expression is placed on the stack.
       Each case has three relevant labels:
         ldo    .. yes, take this case (before block content)
         ldont  .. no, don't take this one (after block content)
         lout   .. we jump here after having taken one block
       We must keep the selector expression on the stack to evaluate the expressions,
       therefore we dup it each time. Before exiting the Select statement, we must drop
       it again. There are two ways of handling this:
       - drop it upon entry into each block
       - drop it after the Select
       For now, use way number one. This generates more code, but doesn't require us
       to provide own Break/Continue/Return statements. */

    /* Parse head */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    if (!tok.checkAdvance("CASE"))
        throw IntError::expectKeyword("Case");
    compileArgumentExpression(bco, scc);
    parseEndOfLine();

    /* Is it valid? */
    if (scc.hasFlags(scc.RefuseBlocks))
        throw IntError::invalidMultiline();

    /* Find first case */
    while (1) {
        cs.readNextLine();
        if (cs.isEOF()) {
            /* End of file */
            throw IntError("Unexpected end of script");
        } else if (tok.getCurrentToken() == tok.tEnd) {
            /* Ok, blank line */
        } else if (tok.checkAdvance("ENDSELECT")) {
            /* Simple quick case */
            parseEndOfLine();
            bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
            return CompiledBlock;
        } else if (tok.checkAdvance("CASE")) {
            /* Start work */
            break;
        } else {
            throw IntError::expectKeyword("Case");
        }
    }

    /* Compile cases. At entrance into this loop, we have parsed the "Case" keyword.
       We exit the loop when seeing EndSelect. */
    IntBytecodeObject::label_t lout = bco.makeLabel();
    while (1) {
        if (tok.checkAdvance("ELSE")) {
            /* Special case, must be last one. */
            parseEndOfLine();
            cs.readNextLine();
            bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
            compileList(bco, IntStatementCompilationContext(&scc).setBlockSyntax().withoutFlags(IntCompilationContext::LinearExecution));
            bco.addJump(IntOpcode::jAlways, lout);
            if (!tok.checkAdvance("ENDSELECT"))
                throw IntError::expectKeyword("EndSelect");
            break;
        } else {
            /* Possibly multi-part condition. */
            IntBytecodeObject::label_t ldo = bco.makeLabel();
            IntBytecodeObject::label_t ldont = bco.makeLabel();
            compileSelectCondition(bco, scc, ldo);
            bco.addJump(IntOpcode::jAlways, ldont);
            bco.addLabel(ldo);
            bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
            compileList(bco, IntStatementCompilationContext(&scc).setBlockSyntax().withoutFlags(IntCompilationContext::LinearExecution));
            bco.addJump(IntOpcode::jAlways, lout);
            bco.addLabel(ldont);
            if (tok.checkAdvance("ENDSELECT"))
                break;
            if (!tok.checkAdvance("CASE"))
                throw IntError::expectKeyword("EndSelect", "Case");
        }
    }
    parseEndOfLine();
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
    bco.addLabel(lout);
    return CompiledBlock;
}

/** Compile "Sub/Function" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context
    \param proc [in] true if this is going to be a procedure (Sub), false if it's a function */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileSub(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, bool proc)
{
    if (scc.hasFlags(scc.RefuseBlocks))
        throw IntError::invalidMultiline();

    /* Read function name */
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();
    if (tok.getCurrentToken() != tok.tIdentifier)
        throw IntError::expectIdentifier(proc ? "subroutine name" : "function name");
    const string_t name = tok.getCurrentString();
    validateName(name);
    tok.readNext();

    /* Create new BCO */
    IntBCORef nbco = new IntBytecodeObject();
    nbco->setIsProcedure(proc);
    nbco->setName(name);
    nbco->setFileName(bco.getFileName());

    /* Read parameters */
    if (tok.checkAdvance(tok.tLParen) && !tok.checkAdvance(tok.tRParen)) {
        bool optional = false;
        while (1) {
            if (tok.checkAdvance("OPTIONAL")) {
                if (optional)
                    throw IntError::misplacedKeyword("Optional");
                optional = true;
            }
            if (tok.getCurrentToken() != tok.tIdentifier)
                throw IntError::expectIdentifier("parameter name");
            validateName(tok.getCurrentString());
            nbco->addArgument(tok.getCurrentString(), optional);
            tok.readNext();

            if (tok.checkAdvance(tok.tRParen))
                break;
            if (!tok.checkAdvance(tok.tComma))
                throw IntError::expectSymbol(",", ")");
        }
    }
    parseEndOfLine();

    /* Header has been read; now read content. Use a new compilation context. */
    cs.readNextLine();
    compileList(*nbco, IntStatementCompilationContext()
                       .setBlockSyntax()
                       .withFlags(IntCompilationContext::LocalContext | IntCompilationContext::LinearExecution));

    /* If it is a function, make sure it returns anything */
    if (!proc)
        nbco->addPushLiteral(0);
    nbco->relocate();

    if (!tok.checkAdvance(proc ? "ENDSUB" : "ENDFUNCTION"))
        throw IntError::expectKeyword(proc ? "EndSub" : "EndFunction");
    tok.checkAdvance(name.c_str());
    parseEndOfLine();

    /* Routine has been compiled. Generate code. */
    bco.addInstruction(IntOpcode::maPush,    IntOpcode::sSubroutine,     bco.addSubroutine(nbco));
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialDefSub, bco.addName(name));

    return CompiledBlock;
}

/** Compile "Try" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileTry(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    IntTokenizer& tok = cs.getTokenizer();

    /* Compilation context */
    struct TryStatementCompilationContext : public IntStatementCompilationContext {
        TryStatementCompilationContext(const IntStatementCompilationContext* parent)
            : IntStatementCompilationContext(parent)
            {
                withoutFlags(LinearExecution);
            }
        void compileContinue(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialUncatch, 0);
                IntStatementCompilationContext::compileContinue(bco);
            }
        void compileBreak(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialUncatch, 0);
                IntStatementCompilationContext::compileBreak(bco);
            }
    };

    /* Make labels */
    IntBytecodeObject::label_t lcatch = bco.makeLabel();
    IntBytecodeObject::label_t lend = bco.makeLabel();

    bco.addJump(IntOpcode::jCatch, lcatch);

    if (tok.readNext() == tok.tEnd) {
        /* Multi-line */
        if (scc.hasFlags(IntCompilationContext::RefuseBlocks))
            throw IntError::invalidMultiline();
        cs.readNextLine();
        compileList(bco, TryStatementCompilationContext(&scc).setBlockSyntax());
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialUncatch, 0);
        if (tok.checkAdvance("ELSE")) {
            bco.addJump(IntOpcode::jAlways, lend);
            bco.addLabel(lcatch);
            bco.addVariableReferenceInstruction(IntOpcode::maPop, "SYSTEM.ERR", scc);
            parseEndOfLine();
            cs.readNextLine();
            compileList(bco, TryStatementCompilationContext(&scc).setBlockSyntax());
            bco.addLabel(lend);
        } else {
            bco.addJump(IntOpcode::jAlways, lend);
            bco.addLabel(lcatch);
            bco.addVariableReferenceInstruction(IntOpcode::maPop, "SYSTEM.ERR", scc);
            bco.addLabel(lend);
        }
        if (!tok.checkAdvance("ENDTRY"))
            throw IntError::expectKeyword("EndTry");
        parseEndOfLine();
        return CompiledBlock;
    } else {
        /* One-liner */
        compile(bco, TryStatementCompilationContext(&scc).setOneLineSyntax());
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialUncatch, 0);
        bco.addJump(IntOpcode::jAlways, lend);
        bco.addLabel(lcatch);
        bco.addVariableReferenceInstruction(IntOpcode::maPop, "SYSTEM.ERR", scc);
        bco.addLabel(lend);
        return CompiledStatement;
    }
}

/** Compile "With" statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileWith(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    IntTokenizer& tok = cs.getTokenizer();
    tok.readNext();

    /* Compilation context */
    struct WithStatementCompilationContext : public IntStatementCompilationContext {
        WithStatementCompilationContext(const IntStatementCompilationContext* parent)
            : IntStatementCompilationContext(parent)
            {
                withoutFlags(LocalContext);
                withExecutionContext(0);
            }
        void compileContinue(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndWith, 0);
                IntStatementCompilationContext::compileContinue(bco);
            }
        void compileBreak(IntBytecodeObject& bco) const
            {
                bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndWith, 0);
                IntStatementCompilationContext::compileBreak(bco);
            }
    };

    /* Expression */
    compileArgumentExpression(bco, scc);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialWith, 0);

    /* Check remaining form */
    bool oneliner = tok.checkAdvance("DO");
    if (tok.getCurrentToken() == tok.tEnd) {
        /* Multi-line */
        if (scc.hasFlags(IntCompilationContext::RefuseBlocks))
            throw IntError::invalidMultiline();
        cs.readNextLine();
        compileList(bco, WithStatementCompilationContext(&scc).setBlockSyntax());
        if (!tok.checkAdvance("ENDWITH"))
            throw IntError::expectKeyword("EndWith");
        parseEndOfLine();
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndWith, 0);
        return CompiledBlock;
    } else {
        /* One line */
        if (!oneliner)
            throw IntError::expectKeyword("Do");
        compile(bco, WithStatementCompilationContext(&scc).setOneLineSyntax());
        bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndWith, 0);
        return CompiledStatement;
    }
}

/** Compile expression statement.
    \param bco [out] bytecode output
    \param scc [in] statement compilation context */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileExpressionStatement(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* Parse expression */
    std::auto_ptr<IntExprNode> node(IntExpressionParser(cs.getTokenizer()).parse());
    if (cs.getTokenizer().getCurrentToken() != IntTokenizer::tEnd)
        throw IntError::garbageAtEnd(true);

    /* If the topmost node is a comparison for equality, compile an assignment instead. */
    if (IntCaseExprNode* cen = dynamic_cast<IntCaseExprNode*>(node.get())) {
        if (IntExprNode* nn = cen->convertToAssignment())
            node.reset(nn);
    }

    /* Compile it */
    if (scc.hasFlags(IntCompilationContext::ExpressionsAreStatements)) {
        node->compileEffect(bco, scc);
        return CompiledStatement;
    } else {
        node->compileValue(bco, scc);
        return CompiledExpression;
    }
}

/** Compile procedure call. Assumes current token being the procedure name.
    \param bco [out] Bytecode goes here
    \param scc [in] Compilation flags */
IntStatementCompiler::StatementResult
IntStatementCompiler::compileProcedureCall(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    /* First is an identifier. */
    IntTokenizer& tok = cs.getTokenizer();
    string_t name = tok.getCurrentString();
    tok.readNext();

    /* Compile args */
    ptr_vector<IntExprNode> args;
    parseArgumentList(args);
    for (ptr_vector<IntExprNode>::size_type i = 0; i != args.size(); ++i)
        args[i]->compileValue(bco, scc);

    /* Call */
    bco.addVariableReferenceInstruction(IntOpcode::maPush, name, scc);
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMProc, args.size());

    return CompiledStatement;
}

/** Compile argument expression. Parse one expression, and compile it
    into a value. Expects current token being first of expression.
    Leaves current token the one after the expression.
    \param bco [out] Bytecode goes here
    \param scc [in] Compilation flags */
void
IntStatementCompiler::compileArgumentExpression(IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    std::auto_ptr<IntExprNode> node(IntExpressionParser(cs.getTokenizer()).parse());
    node->compileValue(bco, scc);
}

/** Compile argument condition. Parse one expression, and compile it
    into a yes/no decision. Expects current token being first of expression.
    Leaves current token the one after the expression.
    \param bco [out] Bytecode goes here
    \param scc [in] Compilation flags
    \param ift [in] If result is true, generate jump here.
    \param iff [in] If result is false (or empty), generate jump here. */
void
IntStatementCompiler::compileArgumentCondition(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff)
{
    std::auto_ptr<IntExprNode> node(IntExpressionParser(cs.getTokenizer()).parse());
    node->compileCondition(bco, scc, ift, iff);
}

/** Compile condition in "Case". Expects current token being the one following "Case".
    Leaves input pointer at end of line (=end of condition).
    \param bco [out] Bytecode goes here
    \param scc [in] Compilation flags
    \param ldo [in] Generated code jumps here if the case matches */
void
IntStatementCompiler::compileSelectCondition(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntBytecodeObject::label_t ldo)
{
    IntTokenizer& tok = cs.getTokenizer();
    while (1) {
        uint8_t relation = biCompareEQ_NC;
        if (tok.checkAdvance("IS")) {
            /* Relation sign plus expression */
            if (tok.checkAdvance(tok.tGT))
                relation = biCompareGT_NC;
            else if (tok.checkAdvance(tok.tGE))
                relation = biCompareGE_NC;
            else if (tok.checkAdvance(tok.tLT))
                relation = biCompareLT_NC;
            else if (tok.checkAdvance(tok.tLE))
                relation = biCompareLE_NC;
            else if (tok.checkAdvance(tok.tNE))
                relation = biCompareNE_NC;
            else if (tok.checkAdvance(tok.tEQ))
                relation = biCompareEQ_NC;
            else
                throw IntError("Expecting relation");
        }
        /* Single expression */
        bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDup, 0);
        compileArgumentExpression(bco, scc);
        bco.addInstruction(IntOpcode::maBinary, relation, 0);
        bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, ldo);

        if (tok.getCurrentToken() == tok.tEnd) {
            break;
        } else if (tok.checkAdvance(tok.tComma)) {
            /* There's another one */
        } else {
            throw IntError::expectSymbol(",");
        }
    }
}

/** Parse argument list. Parses comma-separated list of expressions.
    Terminates successfully when finding end of line. Note that this only
    parses, it does not compile the expressions.
    \param args [out] Arguments expression trees are accumulated here */
void
IntStatementCompiler::parseArgumentList(ptr_vector<IntExprNode>& args)
{
    IntTokenizer& tok = cs.getTokenizer();
    if (tok.getCurrentToken() != tok.tEnd) {
        /* We have some arguments */
        while (1) {
            args.push_back(IntExpressionParser(tok).parse());
            if (tok.getCurrentToken() == tok.tComma) {
                /* Okay, we have one more */
                tok.readNext();
            } else if (tok.getCurrentToken() == tok.tEnd) {
                /* Okay, this was the last one */
                break;
            } else {
                /* Error */
                throw IntError::garbageAtEnd(true);
            }
        }
    }
}

/** Ensure that line ends here. */
void
IntStatementCompiler::parseEndOfLine()
{
    if (cs.getTokenizer().getCurrentToken() != IntTokenizer::tEnd)
        throw IntError::garbageAtEnd(false);
}

/** Validate name. Throws error if the name should not be used. */
void
IntStatementCompiler::validateName(const string_t& name)
{
    // Fundamental functions are permitted as variable and procedure names,
    // because they are only recognized when followed by a "(".
    if (lookupKeyword(name) != kwNone)
        throw IntError("\"" + name + "\" is a reserved name");
}
