/**
  *  \file resmgr/resmgr.cc
  *  \brief Resource Manager -- Interface Implementation
  *
  *  \todo A big problem of this system is lifetime management of the
  *  ResId strings. I definitely do not want a full string_t here for
  *  performance reasons; ResId should remain lightweight. However, we
  *  have the problem that getPixmap() etc clone the ResId and
  *  possibly keep it longer than the actual call. So far, all ResIds
  *  are constructed from static strings, but maybe that changes in
  *  the future. Even then, having users to think about this lifetime
  *  issue is bad. One possible solution would be to make a string
  *  pool which has the same lifetime as the respective cache.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <algorithm>
#include "resmgr/resmgr.h"
#include "cpluslib/assert.h"
#include "util/nls.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "gfx/font.h"
#include "util/misc.h"
#include "util/random.h"
#include "util/except.h"
#include "io/textfile.h"

template<typename T>
static inline const typename T::value_type
nth(const T& container, unsigned count)
{
    typename T::const_iterator i = container.begin();
    std::advance(i, count);
    return *i;
}

/********************************* ResId *********************************/

/** Generalize this resource Id. This modifies the resource Id
    in-place. */
void
ResId::generalize()
{
    if (id2)
        id2 = 0;
    else if (id1)
        id1 = 0;
}

/** Compare for equality. */
bool
ResId::operator==(const ResId& rhs) const
{
    using std::strcmp;
    return (strcmp(name, rhs.name) == 0
            && id1 == rhs.id1
            && id2 == rhs.id2);
}

/** Compare less-than. */
bool
ResId::operator<(const ResId& rhs) const
{
    using std::strcmp;
    if (int i = strcmp(name, rhs.name))
        return i < 0;
    if (id1 < rhs.id1)
        return true;
    if (id1 > rhs.id1)
        return false;
    return id2 < rhs.id2;
}

/** Convert to string. This yields the external representation exposed
    to users. */
string_t
ResId::toString() const
{
    string_t rv = name;
    if (id1)
        rv += format("%d", id1);
    if (id2)
        rv += format("_%d", id2);
    return rv;
}

/** Check whether this resource Id matches the given template.
    \param name name to match (must match exactly)
    \param level level to match. 0=match ResId(name),
           1=match ResId(name,id1), 2=match ResId(name,id1,id2) */
bool
ResId::isElem(const char* name, int level) const
{
    int this_lvl = id2 ? 2 : id1 ? 1 : 0;
    return level == this_lvl
        && std::strcmp(name, this->name) == 0;
}

/****************************** ResProvider ******************************/

/** Detach this resource provider from its manager.
    \see ResManager::removeProvider() */
void
ResProvider::detach()
{
    if (manager)
        manager->removeProvider(*this);
}

/*************************** ResSimpleProvider ***************************/

/*! \class ResSimpleProvider

    This resource provider can be used as a template for implementing
    own resource providers. It does nothing, i.e., provides no
    resources. To make a simple resource provider, you can thus
    override only the methods you need. */
void
ResSimpleProvider::loadPixmap(ResManager& /*mgr*/, ResId /*id*/)
{ }

void
ResSimpleProvider::loadSound(ResManager& /*mgr*/, ResId /*id*/)
{ }

void
ResSimpleProvider::loadAnimation(ResManager& /*mgr*/, ResId /*id*/)
{ }

Ptr<GfxFont>
ResSimpleProvider::getFont(ResId /*id*/)
{
    return 0;
}

Ptr<Stream>
ResSimpleProvider::getStream(ResId /*id*/)
{
    return 0;
}

/************************* ResManager::SoundNode *************************/

ResManager::SoundNode::~SoundNode()
{ }

/******************************* ResManager ******************************/

/// Global resource manager instance.
ResManager ResManager::instance;

/*! \class ResManager

    This class manages a list of ResProvider's, and provides a cache
    for resource items that were already loaded.

    Resources are identified by ResId's. The ResManager automatically
    generalizes a resource Id if a detailed request cannot be
    satisfied.

    Resources are cached, and shared if possible. Inability to satisfy
    a request is cached as well, newly-added resource providers are
    not asked again. Therefore you should first initialize the
    ResManager using addProvider() before you ask it for data.

    Although you can create own instances of this class, you should
    use the prepared instance, see ResManager::getInstance(). */

/** Constructor. */
ResManager::ResManager()
{ }

/** Destructor. */
ResManager::~ResManager()
{
    clearPixmapCache();
    clearSoundCache();
    while (!provider_list.empty()) {
        ASSERT(provider_list.front()->manager == this);
        provider_list.front()->detach();
    }
}

/** Add a resource provider. The provider which was added last will be
    asked first for resource items. A resource provider can only be
    registered to one resource manager; if the provider is already
    registered to someone else, it is detached before. */
void
ResManager::addProvider(ResProvider& provider)
{
    provider.detach();
    provider_list.push_front(&provider);
    provider.manager = this;
}

/** Add a newly allocated provider. This assumes that the provider was
    just allocated with 'new', and it will take over ownership of the
    provider object. Otherwise, this function is identical to
    addProvider() */
void
ResManager::addNewProvider(ResProvider* provider)
{
    addProvider(h.add(provider));
}

/** Remove a resource provider.
    \pre resource provider is registered with us */
void
ResManager::removeProvider(ResProvider& provider)
{
    for (provider_list_t::iterator i = provider_list.begin(); i != provider_list.end(); ++i) {
        if (*i == &provider) {
            provider_list.erase(i);
            provider.manager = 0;
            break;
        }
    }
}

/** Add pixmap associated with resource Id. Intended to be called by
    resource providers in response to their loadPixmap() call. */
void
ResManager::addPixmap(ResId id, Ptr<GfxPixmap> pix)
{
    if (pix) {
        pixmap_cache_t::iterator i = pixmap_cache.find(id);
        if (i == pixmap_cache.end())
            i = safeInsert(pixmap_cache, id, new PixmapNode);
        else if (i->second == 0)
            i->second = new PixmapNode;

        i->second->pixmaps.push_back(pix);
        i->second->pixmap_count++;
    }
}

/** Add sound associated with resource Id. Intended to be called by
    resource providers in response to their loadSound() call. */
void
ResManager::addSound(ResId id, SfxSampleData* snd)
{
    if (snd) {
        sound_cache_t::iterator i = sound_cache.find(id);
        if (i == sound_cache.end())
            i = safeInsert(sound_cache, id, new SoundNode);
        else if (i->second == 0)
            i->second = new SoundNode;

        i->second->sounds.push_back(snd);
        i->second->sound_count++;
    }
}

/** Add animation associated with resource Id. Intended to be called
    by resource providers in response to their loadAnimation()
    call. */
void
ResManager::addAnimation(ResId id, Ptr<GfxAnimation> anim)
{
    if (anim) {
        anim_cache_t::iterator i = anim_cache.find(id);
        if (i == anim_cache.end())
            i = safeInsert(anim_cache, id, new AnimationNode);
        else if (i->second == 0)
            i->second = new AnimationNode;

        i->second->animations.push_back(anim);
        i->second->anim_count++;
    }
}

/** Clear pixmap cache. This forgets all pixmaps currently known. */
void
ResManager::clearPixmapCache()
{
    for (pixmap_cache_t::iterator i = pixmap_cache.begin(); i != pixmap_cache.end(); ++i) {
        if (i->second && ! --i->second->use_count)
            delete i->second;
        i->second = 0;
    }
    pixmap_cache_t().swap(pixmap_cache);
}

/** Clear sound cache. This forgets all sounds currently known, and
    invalidates them.
    \see SfxSample::newGeneration() */
void
ResManager::clearSoundCache()
{
    for (sound_cache_t::iterator i = sound_cache.begin(); i != sound_cache.end(); ++i) {
        if (i->second && ! --i->second->use_count)
            delete i->second;
        i->second = 0;
    }
    sound_cache_t().swap(sound_cache);
    SfxSample::newGeneration();
}

/** Clear animation cache. This forgets all animations currently known. */
void
ResManager::clearAnimationCache()
{
    for (anim_cache_t::iterator i = anim_cache.begin(); i != anim_cache.end(); ++i) {
        if (i->second && ! --i->second->use_count)
            delete i->second;
        i->second = 0;
    }
    anim_cache_t().swap(anim_cache);
}

/** Get pixmap. If there is a pixmap associated with the resource Id,
    return it. If there are multiple pixmaps, return an arbitrarily
    chosen one. If there is no pixmap, returns 0. */
Ptr<GfxPixmap>
ResManager::getPixmap(ResId id)
{
    PixmapNode* n = getPixmapNode(id);
    if (n) {
        if (n->pixmap_count == 0)
            return 0;
        else if (n->pixmap_count == 1)
            return n->pixmaps.front();
        else
            return nth(n->pixmaps, myRandom(n->pixmap_count));
    } else {
        return 0;
    }
}

/** Get pixmap. Like getPixmap(), but never returns null. Throws
    FileOpenException if image not found. */
Ptr<GfxPixmap>
ResManager::getPixmapT(ResId id)
{
    Ptr<GfxPixmap> p = getPixmap(id);
    if (!p)
        throw FileOpenException("g." + id.toString(), _("Resource not found"));
    return p;
}

/** Get sound. If there is a sound associated with the resource Id,
    return it. If there are multiple sounds, return an arbitrarily
    chosen one. If there is no sound, returns 0. */
SfxSample
ResManager::getSound(ResId id)
{
    // We cannot load any sound when we don't have sound output; we
    // don't know where to convert it to.
    if (!isSoundAvailable())
        return 0;

    SoundNode* n = getSoundNode(id);
    if (n) {
        if (n->sound_count == 0)
            return 0;
        else if (n->sound_count == 1)
            return SfxSample(n->sounds.front());
        else
            return SfxSample(nth(n->sounds, myRandom(n->sound_count)));
    } else {
        return 0;
    }
}

/** Get animation. If there is an animation associated with the resource Id,
    return it. If there are multiple animations, return an arbitrarily
    chosen one. If there is no animation, returns 0. */
Ptr<GfxAnimation>
ResManager::getAnimation(ResId id)
{
    AnimationNode* n = getAnimationNode(id);
    if (n) {
        if (n->anim_count == 0)
            return 0;
        else if (n->anim_count == 1)
            return n->animations.front();
        else
            return nth(n->animations, myRandom(n->anim_count));
    } else {
        return 0;
    }
}

/** Get pixmap node for resource Id. First, attempt to load relevant
    data from resource managers. If this fails, attempt to load
    generalized data. */
ResManager::PixmapNode*
ResManager::getPixmapNode(ResId id)
{
    /* already in cache? */
    pixmap_cache_t::iterator i = pixmap_cache.find(id);
    if (i != pixmap_cache.end())
        return i->second;

    /* ask resource providers */
    i = pixmap_cache.insert(pixmap_cache_t::value_type(id, 0)).first;
    for (provider_list_t::iterator j = provider_list.begin(); j != provider_list.end(); ++j)
        (*j)->loadPixmap(*this, id);
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        PixmapNode* n = getPixmapNode(g);
        if (n && !i->second) {
            i->second = n;
            ++n->use_count;
        }
        return i->second;
    }
    return 0;
}

/** Get sound node for resource Id. First, attempt to load relevant
    data from resource managers. If this fails, attempt to load
    generalized data. */
ResManager::SoundNode*
ResManager::getSoundNode(ResId id)
{
    /* already in cache? */
    sound_cache_t::iterator i = sound_cache.find(id);
    if (i != sound_cache.end())
        return i->second;

    /* ask resource providers */
    i = sound_cache.insert(sound_cache_t::value_type(id, 0)).first;
    for (provider_list_t::iterator j = provider_list.begin(); j != provider_list.end(); ++j)
        (*j)->loadSound(*this, id);
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        SoundNode* n = getSoundNode(g);
        if (n && !i->second) {
            i->second = n;
            ++n->use_count;
        }
        return i->second;
    }
    return 0;
}

/** Get animation node for resource Id. First, attempt to load
    relevant data from resource managers. If this fails, attempt to
    load generalized data. */
ResManager::AnimationNode*
ResManager::getAnimationNode(ResId id)
{
    /* already in cache? */
    anim_cache_t::iterator i = anim_cache.find(id);
    if (i != anim_cache.end())
        return i->second;

    /* ask resource providers */
    i = anim_cache.insert(anim_cache_t::value_type(id, 0)).first;
    for (provider_list_t::iterator j = provider_list.begin(); j != provider_list.end(); ++j)
        (*j)->loadAnimation(*this, id);
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        AnimationNode* n = getAnimationNode(g);
        if (n && !i->second) {
            i->second = n;
            ++n->use_count;
        }
        return i->second;
    }
    return 0;
}

/** Get stream. Streams are not "cached" in any way, we just ask all
    the providers whether they know it. */
Ptr<Stream>
ResManager::getStream(ResId id)
{
    for (provider_list_t::iterator i = provider_list.begin(); i != provider_list.end(); ++i)
        if (Ptr<Stream> r = (*i)->getStream(id))
            return r;
    return 0;
}

/** Get font. Streams are not cached in any way, we just ask all the
    providers whether they know it. */
Ptr<GfxFont>
ResManager::getFont(ResId id)
{
    for (provider_list_t::iterator i = provider_list.begin(); i != provider_list.end(); ++i)
        if (Ptr<GfxFont> r = (*i)->getFont(id))
            return r;
    return 0;
}

/****************************** ResDirectory *****************************/

static const char* const gfx_suffix_list[] = { ".cd", ".gfx", ".bmp", ".jpg", ".png", 0 };
static const char* const sfx_suffix_list[] = { ".wav", 0 };
static const char* const fnt_suffix_list[] = { ".fnt", 0 };
static const char* const no_suffix_list[] = { 0 };

ResDirectory::alias_map_t ResDirectory::alias_map;

/*! \class ResDirectory

    This manages resources stored in individual files. It provides
    management of alias definitions (index.txt) as well as heuristics
    for guessing file names from resource Ids: a resource "foo.bar" is
    looked up in file "foo/bar.EXT", where EXT is an extension
    matching the resource type. */

/** Constructor. */
ResDirectory::ResDirectory(Ptr<Directory> dir)
    : dir(dir)
{
    Ptr<Stream> s = dir->openFileNT("index.txt", Stream::C_READ);
    if (s) {
        loadAliases(*s);
    }
}

ResDirectory::~ResDirectory()
{ }

void
ResDirectory::loadPixmap(ResManager& mgr, ResId id)
{
    Ptr<GfxPixmap> pix;
    alias_map_t::iterator i = alias_map.find("g." + id.toString());
    if (i != alias_map.end()) {
        pix = loadImage(dir->openFileNT(i->second.name, Stream::C_READ));
    } else {
        pix = loadImage(openStream(id.toString(), gfx_suffix_list));
    }
    if (pix)
        mgr.addPixmap(id, pix);
}

void
ResDirectory::loadSound(ResManager& mgr, ResId id)
{
    alias_map_t::iterator i = alias_map.find("s." + id.toString());

    Ptr<Stream> st =
        (i != alias_map.end()
         ? dir->openFileNT(i->second.name, Stream::C_READ)
         : openStream(id.toString(), sfx_suffix_list));

    if (st) {
        try {
            mgr.addSound(id, new SfxSampleData(*st));
        }
        catch (...) { }
    }
}

void
ResDirectory::loadAnimation(ResManager& /*mgr*/, ResId /*id*/)
{ }

Ptr<GfxFont>
ResDirectory::getFont(ResId id)
{
    Ptr<Stream> s;
    alias_map_t::iterator i = alias_map.find("f." + id.toString());
    if (i != alias_map.end()) {
        s = dir->openFileNT(i->second.name, Stream::C_READ);
    } else {
        s = openStream(id.toString(), fnt_suffix_list);
    }
    if (s)
        return new GfxStdFont(s);
    else
        return 0;
}

Ptr<Stream>
ResDirectory::getStream(ResId id)
{
    alias_map_t::iterator i = alias_map.find(id.toString());
    if (i != alias_map.end()) {
        return dir->openFileNT(i->second.name, Stream::C_READ);
    } else {
        return openStream(id.toString(), no_suffix_list);
    }
}

/** Attempt to open a file generated from a resource name. Generates
    the name by substituting directory separators for the dots and
    trying a couple of extensions. For example, "vcr.banner" is looked
    up as "vcr/banner.png", "vcr/banner.jpg", etc.
    \param name    resource name
    \param suffixes null-terminated list of suffixes
    \return stream, or 0 if none found */
Ptr<Stream>
ResDirectory::openStream(string_t name, const char*const* suffixes)
{
    string_t::size_type n = 0;
    while ((n = name.find('.', n)) != string_t::npos)
        name[n] = Platform::PATH_SEPARATOR, ++n;

    while (*suffixes) {
        if (Ptr<Stream> st = dir->openFileNT(name + *suffixes, Stream::C_READ))
            return st;
        ++suffixes;
    }
    return 0;
}

/** Load alias table from given stream. */
void
ResDirectory::loadAliases(Stream& in)
{
    // FIXME: this does not yet support what it is supposed
    // to support, namely multiple pixmaps per name etc.
    TextFile tf(in);
    string_t line, l, r;
    while (tf.getLine(line)) {
        /* strip comment */
        string_t::size_type p = line.find_first_of(";#");
        if (p != line.npos)
            line.erase(p);
        line = strTrim(line);
        if (line.empty())
            continue;

        /* must be "normal" line */
        if (strSplit(line, l, r, "=")) {
            alias_map[strTrim(l)].name = strTrim(r);
        } else {
            console.write(LOG_ERROR, format(_("%s:%d: file format error -- line ignored"),
                                            in.getName(), tf.getLineNr()));
        }
    }
}
