%
%   PCC2 Standard Library
%
%   This file defines commands and keys used internally by PCC2,
%   and a number of user-callable routines described in the
%   manual. Do not modify unless you are absolutely 100% positively
%   sure what you are doing.
%
%   This file will be overwritten by PCC2 upgrades. It will most
%   likely only work with the PCC2 version it was intended for.
%
%   This version of core.q is for PCC2, and is incompatible with
%   PCC 1.x - and vice versa.
%

%Option Optimize(2)

Dim Shared CC$LibraryVersion = '1.99.9'

%%% Console-Mode Replacements for GUI routines %%%%%%%%%%%%%%%%%%%%%%

If Not System.GUI Then

  %%% Utilities

  % Return the index of the first word in list which starts with the given word.
  Function CC$MatchWord (list, word)
    Local i:=1, e
    Do While list<>''
      e := First(list, ' ')
      If Mid(e,1,Len(word)) = word Then Return i
      i := i + 1
      list := Trim(Rest(list, ' '))
    Loop
  EndFunction

  % Count number of words in list
  Function CC$CountWords (list)
    Local i:=0
    Do While list<>''
      i:=i+1
      list:=Trim(Rest(list, ' '))
    Loop
    Return i
  EndFunction


  %%% Actual User Functions

  % MessageBox generates a console message.
  Sub MessageBox (text, Optional heading)
    % FIXME: PCC 1.x replaces \r by \n in text.
    If IsEmpty(heading) Then heading := Translate("Message")
    Print heading, ": ", text
  EndSub

  % UI.Update has no effect in console mode.
  Sub UI.Update (Optional flag)
  EndSub

  % UI.ChooseObject generates an error in console mode.
  Sub UI.ChooseObject (screen)
    Abort "Not in graphics mode"
  EndSub

  % UI.InputNumber generates an error in console mode (like it does in PCC 1.x,
  % although we can probably do better).
  Sub UI.InputNumber (prompt, Optional lo, hi, cur, help)
    Abort "Not in graphics mode"
  EndSub

  % UI.Message generates a console message and a prompt
  Sub UI.Message (text, Optional heading, buttons)
    MessageBox text, Heading
    If IsEmpty(buttons) Then buttons := Translate("OK")
    Do
      UI.Input buttons
      If IsEmpty(UI.Result) Then
        % This means the last choice
        UI.Result := Z(CC$CountWords(buttons))
        Break
      Else If UI.Result = "" Then
        % This means the first choice
        UI.Result := 1
        Break
      Else
        % Does this match any button?
        UI.Result := CC$MatchWord(buttons, UI.Result)
        If Not IsEmpty(UI.Result) Then Break
      EndIf
    Loop
  EndSub
EndIf


%%% User-callable Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Sub WaitOneTurn ()
% Sub MoveTo (x, y)
% Sub MoveTowards (x, y)
% Sub EnqueueShip (h, e, Optional bt, bc, tt, tc)
% Sub CargoUploadWait (cargo, Optional flags)
% Sub CargoTransferWait (sid, cargo, Optional flags)
% Sub BuildMinesWait (amount)
% Sub BuildFactoriesWait (amount)
% Sub BuildDefenseWait (amount)
% Sub BuildBaseDefenseWait (amount)
% Sub SetFleetName (name)
% Sub ChangeFleetLeader (fid, sid)
% Sub SelectionLayer (l)
% Sub FixShipWait (Optional id)
% Sub RecycleShipWait (Optional id)
% Sub Notify (s, Optional t)
% Sub CC$ScriptError (s)
% Sub AutoLoad (file, name, Optional args)

If System.GUI Then
%% When the GUI is not active, don't define all these to save memory

Dim Shared CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC

% Ctrl-P on starbase/ship build screen
Sub CC$PickupBuild
  If Build.YesNo Then
    CC$BuildH  := Build.Hull$
    CC$BuildE  := Build.Engine$
    CC$BuildBT := Build.Beam$
    CC$BuildBC := Build.Beam.Count
    CC$BuildTT := Build.Torp$
    CC$BuildTC := Build.Torp.Count
    Messagebox Format(Translate("Ship specification for a %s saved. Use [Ctrl-B] on other starbases to build this ship there."), Build.Short)
  EndIf
EndSub

% Ctrl-B on starbase screen
Sub CC$PerformBuild
  Local System.Err
  If IsEmpty(CC$BuildH) Or Owner$<>My.Race$ Then Return
  If Build.YesNo Then
    UI.Message Format(Translate("This starbase is already building a %s. Override this with the saved build order for a %s?"), Build.Short, Hull(CC$BuildH).Name.Short), Translate("Ship Building"), Translate("Yes No")
    If UI.Result<>1 Then Return
  EndIf
  Try
    BuildShip CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC
  Else
    Messagebox Format(Translate("Unable to build a %s: %s"), Hull(CC$BuildH).Name.Short, System.Err)
  EndTry
EndSub

% Sub CC$SearchBuild

% '.' everywhere
Sub CC$ToggleSelection
  Mark Not Marked
EndSub

% F9 on ship
Sub CC$SetShipComment
  UI.Input Translate("Enter new comment for this ship:"), Translate("Edit Comment"), 255, "m30", Comment
  SetComment UI.Result
EndSub

% F9 on planet
Sub CC$SetPlanetComment
  UI.Input Translate("Enter new comment for this planet:"), Translate("Edit Comment"), 255, "m30", Comment
  SetComment UI.Result
EndSub

% Shift-F9 on control screens
Sub CC$SetPlanetCommentAt(X,Y)
  Local pid = PlanetAt(X,Y,1)
  If pid Then With Planet(pid) Do CC$SetPlanetComment
EndSub

% N in ship screen
Sub CC$ShipName
  If Owner$<>My.Race$ Then Return
  UI.Input Translate("Enter new name:"), Format(Translate("Rename Starship #%d"), Id), 20, "gm20", Name
  SetName UI.Result
EndSub

% Sub CC$FleetName
% Sub CC$ChangeLeader

% C-z
Sub CC$SetWaypoint (x, y)
  % PCC 1.x:
  % Try With Lock("s" & Id & ".waypoint") Do SetWaypoint x, y
  Try SetWaypoint x, y
EndSub

EndIf % System.GUI

%%% Initialisation of the Runtime System %%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Properties.
CreateShipProperty   Comment
CreatePlanetProperty Comment

% Keymaps. Note that those differ from PCC 1.x because we support multiple inheritance.
CreateKeymap Global, Ship, Planet, Base, Fleet

CreateKeymap ControlScreen(Global)
CreateKeymap ShipScreen(Ship, ControlScreen)
CreateKeymap PlanetScreen(Planet, ControlScreen)
CreateKeymap BaseScreen(Base, ControlScreen)
CreateKeymap FleetScreen(Fleet, ControlScreen)
CreateKeymap HistoryScreen(ControlScreen)
CreateKeymap AutoTaskScreen(ControlScreen)

CreateKeymap Starchart(Global)
CreateKeymap ShipLock(Ship, Starchart)
CreateKeymap PlanetLock(Planet, Starchart)
CreateKeymap BaseLock(Base, Starchart)

CreateKeymap RaceScreen(Global)
CreateKeymap ShipBuildScreen(Global)

% Global Variables (System.Err is referenced by the interpreter core and therefore defined there).
Dim Shared UI.Result
Dim Shared UI.Directory
Dim Shared UI.Prefix
Dim Shared Cargo.Remainder
Dim Shared Build.Remainder

% Keybindings
Bind Base            "a"      := "CC$BuildAmmo"
Bind Base            "b"      := "CC$BuildShip", "tab" := "CC$BuildShip"
Bind Base            "c-p"    := "CC$PickupBuild", "c-b" := "CC$PerformBuild"
Bind Base            "d"      := "CC$BuildStructures 2"
Bind Base            "f"      := "CC$ChangeFCode"
Bind Base            "m"      := "CC$ChangeMission"
Bind Base            "t"      := "CC$ChangeTech"
Bind Planet          "b"      := "CC$BuildStructures 0"
Bind Planet          "c-s"    := "CC$BuySupplies"
Bind Planet          "f"      := "CC$ChangeFCode"
Bind Planet          "f9"     := "CC$SetPlanetComment"
Bind Planet          "g"      := "CC$Give"
Bind Planet          "s"      := "CC$SellSupplies"
Bind Planet          "t"      := "CC$ChangeTaxes"
Bind PlanetScreen    "d"      := "CC$BuildStructures 2"
Bind PlanetScreen    "m"      := "CC$BuildStructures 1"
Bind PlanetScreen    "tab"    := "CC$BuildStructures 0"
Bind Ship            "c-f"    := "Try SetFCode Planet(Orbit$).FCode"
Bind Ship            "c-z"    := "CC$SetWaypoint Loc.X, Loc.Y"
Bind Ship            "e"      := "CC$ChangePE"
Bind Ship            "f"      := "CC$ChangeFCode"
Bind Ship            "f9"     := "CC$SetShipComment"
Bind Ship            "g"      := "CC$Give"
Bind Ship            "m"      := "CC$ChangeMission"
Bind Ship            "n"      := "CC$ShipName"
Bind Ship            "s"      := "CC$ShipSpec"
Bind ControlScreen   "."      := "CC$ToggleSelection"
Bind ControlScreen   "pgup"   := "CC$CSBrowse 0,0", "-" := "CC$CSBrowse 0,0", "wheelup" := "CC$CSBrowse 0,0"
Bind ControlScreen   "pgdn"   := "CC$CSBrowse 1,0", "+" := "CC$CSBrowse 1,0", "wheeldn" := "CC$CSBrowse 1,0"
Bind ControlScreen   "c-pgup" := "CC$CSBrowse 0,1", "c--" := "CC$CSBrowse 0,1", "c-wheelup" := "CC$CSBrowse 0,1"
Bind ControlScreen   "c-pgdn" := "CC$CSBrowse 1,1", "c-+" := "CC$CSBrowse 1,1", "c-wheeldn" := "CC$CSBrowse 1,1"
Bind ControlScreen   "s-f9"   := "CC$SetPlanetCommentAt Loc.X,Loc.Y"
Bind ShipLock        "."      := "CC$ToggleSelection"
Bind PlanetLock      "."      := "CC$ToggleSelection"
Bind Starchart       "s-f9"   := "CC$SetPlanetCommentAt Loc.X,Loc.Y"

On EnterDirectory Do TryLoad "autoexec.q"

If System.GUI Then Print "[", System.Program, " ", System.Version, ", core.q ", CC$LibraryVersion, "]"
