/**
  *  \file u/t_int_statement.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/statement.h"

#include "u/t_int.h"
#include "int/exec.h"
#include "int/csource.h"
#include "int/interror.h"

#define gva globalVars[0]
#define gvb globalVars[1]
#define gvc globalVars[2]

/** Test expression statement. Verifies that
    - the statement compiles into an expression statement
    - runs correctly
    - yields the expected value
    \param expr  Expression statement
    \param value Expected result value */
static void
testIntExprStatement(const char* expr, int value)
{
    try {
        IntMemoryCommandSource mcs(expr);

        IntExecutionContext exec;
        exec.pushNewContext(new GlobalTestVars());

        IntStatementCompilationContext scc;
        scc.withExecutionContext(&exec);
        scc.withFlags(IntCompilationContext::RefuseBlocks);
        scc.withFlags(IntCompilationContext::LinearExecution);

        IntBCORef bco = new IntBytecodeObject();
        IntStatementCompiler::StatementResult result = IntStatementCompiler(mcs).compile(*bco, scc);

        TSM_ASSERT_EQUALS(expr, result, IntStatementCompiler::CompiledExpression);

        exec.pushNewFrame(new IntExecutionFrame(bco));
        exec.run();

        IntValue* v = exec.getResult();
        TSM_ASSERT(expr, v);

        IntIntValue* iv = dynamic_cast<IntIntValue*>(v);
        TSM_ASSERT(expr, iv);
        TSM_ASSERT_EQUALS(expr, iv->getValue(), value);
    }
    catch (IntError& e) {
        printf("exception: %s\n", e.what());
        TSM_ASSERT(expr, 0);
    }
    catch (...) {
        TSM_ASSERT(expr, 0);
    }
}

/** Test statements. Given a (possibly multi-line) statement, Verifies that
    - the statements compile into anything but an expression statement
    - run correctly
    \param stmt Statements, separated by '\n' */
static void
testStatement(const char* stmt)
{
    try {
        // Build a command source
        IntMemoryCommandSource mcs;
        const char* q = stmt;
        while (const char* p = strchr(q, '\n')) {
            mcs.addLine(string_t(q, p - q));
            q = p+1;
        }
        mcs.addLine(q);

        // Build compilation environment
        IntExecutionContext exec;
        exec.pushNewContext(new GlobalTestVars());

        IntStatementCompilationContext scc;
        scc.withExecutionContext(&exec);
        scc.withFlags(IntCompilationContext::LinearExecution);
        scc.withFlags(IntCompilationContext::ExpressionsAreStatements);

        IntBCORef bco = new IntBytecodeObject();
        IntStatementCompiler::StatementResult result = IntStatementCompiler(mcs).compileList(*bco, scc);

        TSM_ASSERT_DIFFERS(stmt, result, IntStatementCompiler::CompiledExpression);

        exec.pushNewFrame(new IntExecutionFrame(bco));
        exec.run();
    }
    catch (IntError& e) {
        printf("exception: %s\n", e.what());
        TSM_ASSERT(stmt, 0);
    }
    catch (...) {
        TSM_ASSERT(stmt, 0);
    }
}

void
IntStatementTestSuite::setUp()
{
    gva = gvb = gvc = 0;
}

void
IntStatementTestSuite::tearDown()
{ }

/** Test expression statements. This tests just the parser. The
    expression interpreter is tested in detail in t_int_parseexpr.cc,
    therefore the expressions can be simple, and we limit ourselves to
    testing expressions yielding integers. The idea is to simply make
    sure that we correctly compile syntactically ambiguous statements. */
void
IntStatementTestSuite::testExprStatement()
{
    // Operators: ";"
    testIntExprStatement("a;97", 97);

    // Operators: ":="
    testIntExprStatement("a:=3", 3);
    testIntExprStatement("b:=c:=0", 0);

    // Operators: "Or", "Xor"
    testIntExprStatement("a or b", 1);
    testIntExprStatement("a xor a", 0);

    // Operators: "And"
    testIntExprStatement("a and a", 1);

    // Operators: "Not"
    testIntExprStatement("not a", 0);

    // Operators: comparisons
    testIntExprStatement("a>0", 1);
    testIntExprStatement("a<10", 1);
    testIntExprStatement("a<3", 0);
    testIntExprStatement("a>=0", 1);
    testIntExprStatement("a<=3", 1);
    testIntExprStatement("a<>99", 1);
    testIntExprStatement("a=7", 7);               // assignment
    testIntExprStatement("a=3 or 2", 1);          // comparison

    // Operators: "#", "&"
    testIntExprStatement("a&b;9", 9);
    testIntExprStatement("a#b;9", 9);

    // Operators: "+", "-"
    testIntExprStatement("a+3", 10);
    testIntExprStatement("a-3", 4);

    // Operators: "*", "/", "\", "Mod"
    testIntExprStatement("a*3", 21);
    testIntExprStatement("a/1;12", 12);
    testIntExprStatement("a\\2", 3);
    testIntExprStatement("a mod 2", 1);

    // Operators: unary "+", "-"
    testIntExprStatement("-3", -3);
    testIntExprStatement("+3", +3);

    // Operators: "^"
    testIntExprStatement("a^2", 49);

    // Operators: "(...)"
    testIntExprStatement("(9)", 9);
    testIntExprStatement("(9)*2", 18);

    // Operators: function call
    testIntExprStatement("isempty(z(0))", 1);

    // Firsts: identifiers
    testIntExprStatement("a", 7);

    // Firsts: numbers
    testIntExprStatement("1+1", 2);
    testIntExprStatement("1.3*99;5", 5);

    // Firsts: strings
    testIntExprStatement("'a';99", 99);
    testIntExprStatement("'a'+'b';98", 98);
}


/** Test various flavours of "If", "Else", "Else If". */
void
IntStatementTestSuite::testIf()
{
    testStatement("a:=3");
    TS_ASSERT_EQUALS(gva, 3);

    testStatement("if a=4 then\n"
                  "  a:=5\n"
                  "else\n"
                  "  a:=6\n"
                  "endif");
    TS_ASSERT_EQUALS(gva, 6);

    testStatement("if a=5 then\n"
                  "  a:=6\n"
                  "else if a=6 then\n"
                  "  a=7\n"
                  "endif");
    TS_ASSERT_EQUALS(gva, 7);

    testStatement("if a=5 then\n"
                  "  a:=6\n"
                  "else if a=6 then\n"
                  "  a=7\n"
                  "else if a=7 then\n"
                  "  a=8\n"
                  "else\n"
                  "  a:=9\n"
                  "endif");
    TS_ASSERT_EQUALS(gva, 8);

    testStatement("if a=5 then %second\n"
                  "  a:=6\n"
                  "else if a=6 then\n"
                  "  a=7\n"
                  "else if a=7 then\n"
                  "  a=8\n"
                  "else\n"
                  "  a:=9\n"
                  "endif");
    TS_ASSERT_EQUALS(gva, 9);

    testStatement("if a=9 then a:=10");
    TS_ASSERT_EQUALS(gva, 10);
}

/** Test 'For' statement. */
void
IntStatementTestSuite::testFor()
{
    // Basic iteration
    testStatement("for b:=1 to 10 do a:=a+b");
    TS_ASSERT_EQUALS(gva, 55);

    // Backward iteration: body must not be entered
    testStatement("for b:=10 to 1 do abort 1");

    // Body must be entered once
    gva = 0;
    testStatement("for b:=20 to 20 do a:=a+99");
    TS_ASSERT_EQUALS(gva, 99);

    gva = 0;
    testStatement("for b:=-20 to -20 do a:=a+b");
    TS_ASSERT_EQUALS(gva, -20);

    // Basic iteration, multi-line
    testStatement("a:=0\n"
                  "for b:=1 to 10\n"
                  "  a:=a+b\n"
                  "next");
    TS_ASSERT_EQUALS(gva, 55);

    // Basic iteration, multi-line, optional 'do' keyword
    testStatement("a:=10\n"
                  "for b:=1 to 10 do\n"
                  "  a:=a+b\n"
                  "next");
    TS_ASSERT_EQUALS(gva, 65);

    // Continue
    testStatement("a:=0\n"
                  "for b:=1 to 10 do\n"
                  "  if b mod 2 = 0 then continue\n"
                  "  a:=a+b\n"
                  "next");
    TS_ASSERT_EQUALS(gva, 25);

    // Break
    testStatement("a:=0\n"
                  "for b:=1 to 10 do\n"
                  "  if b mod 2 = 0 then break\n"
                  "  a:=a+b\n"
                  "next");
    TS_ASSERT_EQUALS(gva, 1);

    // Varying limit (must not affect loop)
    testStatement("c:=10; a:=0\n"
                  "for b:=1 to c do\n"
                  "  c:=3\n"
                  "  a:=a+1\n"
                  "next");
    TS_ASSERT_EQUALS(gvc, 3);
    TS_ASSERT_EQUALS(gva, 10);
}

/** Test "Do"/"Loop" statements. */
void
IntStatementTestSuite::testDo()
{
    // Basic Do/While loop
    testStatement("a:=1; b:=0\n"
                  "do while a<10\n"
                  "  a:=a+1\n"
                  "  b:=b+1\n"
                  "loop");
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 9);

    // Basic Do/Until loop
    testStatement("a:=1; b:=0\n"
                  "do until a>10\n"
                  "  a:=a+1\n"
                  "  b:=b+1\n"
                  "loop");
    TS_ASSERT_EQUALS(gva, 11);
    TS_ASSERT_EQUALS(gvb, 10);

    // Basic Do/Loop/While loop
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  a:=a+1\n"
                  "  b:=b+1\n"
                  "loop while a<10");
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 9);

    // Basic Do/Loop/Until loop
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  a:=a+1\n"
                  "  b:=b+1\n"
                  "loop until a>10");
    TS_ASSERT_EQUALS(gva, 11);
    TS_ASSERT_EQUALS(gvb, 10);

    // Do/While entered with wrong condition
    testStatement("a:=1; b:=0\n"
                  "do while a<1\n"
                  "  b:=99\n"
                  "loop");
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 0);

    // Do/Loop/While entered with wrong condition
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  b:=b+99\n"
                  "loop while a<1");
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 99);

    // Condition with side-effect
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  b:=b+1\n"
                  "loop while (a:=a+1)<10");
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 9);

    // Continue
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  continue\n"
                  "  b:=b+1\n"
                  "loop while (a:=a+1)<10");
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 0);

    // Break
    testStatement("a:=1; b:=0\n"
                  "do\n"
                  "  break\n"
                  "  b:=b+1\n"
                  "loop while (a:=a+1)<10");
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 0);
}

/** Test "Select Case" statements. */
void
IntStatementTestSuite::testSelect()
{
    // Basic Select Case
    gva = 1;
    testStatement("select case a\n"
                  "  case 0\n"
                  "    b:=9\n"
                  "  case 1\n"
                  "    b:=8\n"
                  "  case 2\n"
                  "    b:=7\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 8);

    // No matching case
    gva = 1;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 10\n"
                  "    b:=9\n"
                  "  case 11\n"
                  "    b:=8\n"
                  "  case 12\n"
                  "    b:=7\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 0);

    // No matching case, matching else
    gva = 1;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 10\n"
                  "    b:=9\n"
                  "  case 11\n"
                  "    b:=8\n"
                  "  case 12\n"
                  "    b:=7\n"
                  "  case else\n"
                  "    b:=6\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 6);

    // Ranges
    gva = 5;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 1,2,3\n"
                  "    b:=1\n"
                  "  case 4,5,6\n"
                  "    b:=2\n"
                  "  case 7,8,9\n"
                  "    b:=3\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 2);

    // Match first in range
    gva = 1;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 1,2,3\n"
                  "    b:=1\n"
                  "  case 4,5,6\n"
                  "    b:=2\n"
                  "  case 7,8,9\n"
                  "    b:=3\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 1);

    // Match last in range
    gva = 3;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 1,2,3\n"
                  "    b:=1\n"
                  "  case 4,5,6\n"
                  "    b:=2\n"
                  "  case 7,8,9\n"
                  "    b:=3\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 1);

    // Match last item
    gva = 9;
    gvb = 0;
    testStatement("select case a\n"
                  "  case 1,2,3\n"
                  "    b:=1\n"
                  "  case 4,5,6\n"
                  "    b:=2\n"
                  "  case 7,8,9\n"
                  "    b:=3\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 3);

    // Relations
    gva = 5;
    gvb = 0;
    testStatement("select case a\n"
                  "  case is <5\n"
                  "    b:=1\n"
                  "  case is >=5\n"
                  "    b:=2\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 2);

    // Empty
    gva = 0;
    gvb = 0;
    testStatement("select case a\n"
                  "endselect");

    // Empty with Else
    gva = 0;
    gvb = 0;
    testStatement("select case a\n"
                  "  case else\n"
                  "    b:=3\n"
                  "endselect");
    TS_ASSERT_EQUALS(gvb, 3);

    // Break from switch (interesting because both for and select place stuff on the stack)
    gva = 0;
    gvb = 0;
    testStatement("for c:=1 to 10 do\n"
                  "  select case c\n"
                  "    case 1,3,5,7,9\n"
                  "      b:=b+c\n"            // 1,3,5,7
                  "    case is <5\n"
                  "      b:=b+2*c\n"          // 2*2, 2*4
                  "    case is =8\n"
                  "      break\n"
                  "    case else\n"           // 3*6
                  "      b:=b+3*c\n"
                  "  endselect\n"
                  "next");
    TS_ASSERT_EQUALS(gvb, 46);

    // Continue from switch
    gva = 0;
    gvb = 0;
    testStatement("for c:=1 to 10 do\n"
                  "  select case c\n"
                  "    case 1,3,5,7,9\n"
                  "      b:=b+c\n"            // 1,3,5,7,9
                  "    case is =8\n"
                  "      continue\n"
                  "  endselect\n"
                  "  b:=b+1\n"
                  "next");
    TS_ASSERT_EQUALS(gvb, 34);
}

/** Test Eval statement. */
void
IntStatementTestSuite::testEval()
{
    testStatement("Eval 'a:=1'");
    TS_ASSERT_EQUALS(gva, 1);

    testStatement("Eval 'a:=2', 'b:=a+3'");
    TS_ASSERT_EQUALS(gva, 2);
    TS_ASSERT_EQUALS(gvb, 5);

    testStatement("Eval 'for a:=1 to 10', 'b:=a+5', 'next'");
    TS_ASSERT_EQUALS(gvb, 15);
}
